"""This module implements hypothesis strategies for BQSKit."""
from __future__ import annotations

import inspect
from typing import Any
from typing import Sequence

from hypothesis.control import assume
from hypothesis.strategies import composite
from hypothesis.strategies import deferred
from hypothesis.strategies import dictionaries
from hypothesis.strategies import floats
from hypothesis.strategies import from_type
from hypothesis.strategies import integers
from hypothesis.strategies import just
from hypothesis.strategies import lists
from hypothesis.strategies import one_of
from hypothesis.strategies import permutations
from hypothesis.strategies import sampled_from
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import text
from hypothesis.strategies import tuples

import bqskit.ir.gates
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.gates.composed.daggergate import DaggerGate
from bqskit.ir.gates.composed.frozenparam import FrozenParameterGate
from bqskit.ir.gates.composed.tagged import TaggedGate
from bqskit.ir.gates.constant.identity import IdentityGate
from bqskit.ir.gates.constant.unitary import ConstantUnitaryGate
from bqskit.ir.gates.parameterized.pauli import PauliGate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.ir.interval import CycleInterval
from bqskit.ir.location import CircuitLocation
from bqskit.ir.location import CircuitLocationLike
from bqskit.ir.operation import Operation
from bqskit.ir.point import CircuitPoint
from bqskit.ir.point import CircuitPointLike
from bqskit.ir.region import CircuitRegion
from bqskit.qis.state.state import StateLike
from bqskit.qis.state.state import StateVector
from bqskit.qis.unitary import RealVector
from bqskit.qis.unitary import UnitaryMatrix
from bqskit.qis.unitary.unitarymatrix import UnitaryLike
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_sequence_of_int


@composite
def num_qudits(
    draw: Any,
    max_num_qudits: int = 8,
    min_num_qudits: int = 1,  # TODO: Param order should be swapped
) -> int:
    """Hypothesis strategy for generating a valid number of qudits."""
    return draw(integers(min_num_qudits, max_num_qudits))


@composite
def radixes(
    draw: Any,
    max_num_qudits: int = 4,
    allowed_bases: Sequence[int] = (2, 3, 4),
    min_num_qudits: int = 1,
) -> tuple[int, ...]:
    """Hypothesis strategy for generating a valid radixes object."""
    num_q = draw(num_qudits(max_num_qudits, min_num_qudits))
    x = [sampled_from(allowed_bases) for _ in range(num_q)]
    return draw(tuples(*x))


@composite
def num_qudits_and_radixes(
    draw: Any,
    max_num_qudits: int = 4,
    allowed_bases: Sequence[int] = (2, 3, 4),
    min_num_qudits: int = 1,
) -> tuple[int, tuple[int, ...]]:
    """Hypothesis strategy for a matching pair of num_qudits and radixes."""
    num_q = draw(num_qudits(max_num_qudits, min_num_qudits))
    x = [sampled_from(allowed_bases) for _ in range(num_q)]
    return (num_q, draw(tuples(*x)))


@composite
def unitaries(
    draw: Any,
    max_num_qudits: int = 3,
    allowed_bases: Sequence[int] = (2, 3),
    min_num_qudits: int = 1,
) -> UnitaryMatrix:
    """Hypothesis strategy for generating `UnitaryMatrix`'s."""
    num_qudits, radixes = draw(
        num_qudits_and_radixes(
            max_num_qudits, allowed_bases, min_num_qudits,
        ),
    )
    return UnitaryMatrix.random(num_qudits, radixes)


@composite
def unitary_likes(
    draw: Any,
    max_num_qudits: int = 3,
    allowed_bases: Sequence[int] = (2, 3),
    min_num_qudits: int = 1,
) -> UnitaryLike:
    """Hypothesis strategy for generating UnitaryLike objects."""
    utry = draw(unitaries(max_num_qudits, allowed_bases, min_num_qudits))
    return draw(sampled_from([utry, utry.numpy]))


@composite
def state_vectors(
    draw: Any,
    max_num_qudits: int = 3,
    allowed_bases: Sequence[int] = (2, 3),
    min_num_qudits: int = 1,
) -> StateVector:
    """Hypothesis strategy for generating `StateVector`'s."""
    num_qudits, radixes = draw(
        num_qudits_and_radixes(
            max_num_qudits, allowed_bases, min_num_qudits,
        ),
    )
    return StateVector.random(num_qudits, radixes)


@composite
def state_likes(
    draw: Any,
    max_num_qudits: int = 3,
    allowed_bases: Sequence[int] = (2, 3),
    min_num_qudits: int = 1,
) -> StateLike:
    """Hypothesis strategy for generating StateLike objects."""
    vec = draw(state_vectors(max_num_qudits, allowed_bases, min_num_qudits))
    return draw(sampled_from([vec, vec.numpy]))


gate_instances: list[Gate] = []
for gate_class_str in bqskit.ir.gates.__all__:
    gate_class = getattr(bqskit.ir.gates, gate_class_str)
    gate_params = inspect.signature(gate_class.__init__).parameters
    if (
        len(gate_params) == 3
        and 'self' in gate_params
        and 'args' in gate_params
        and 'kwargs' in gate_params
        and not inspect.isabstract(gate_class)
        and not gate_class.__name__ == 'ConstantGate'
    ):
        gate_instances.append(gate_class())


@composite
def constant_unitary_gates(
    draw: Any,
    radixes: Sequence[int],
) -> ConstantUnitaryGate:
    """Hypothesis strategy for generating `ConstantUnitaryGate`'s."""
    utry = draw(unitaries(len(radixes), radixes, len(radixes)))
    return ConstantUnitaryGate(utry, radixes)


@composite
def dagger_gates(
    draw: Any,
    radixes: Sequence[int] | int | None = None,
    constant: bool = False,
) -> DaggerGate:
    """Hypothesis strategy for generating `DaggerGate`'s."""
    gate = draw(deferred(lambda: gates(radixes, constant)))
    return DaggerGate(gate)


@composite
def frozen_gates(
    draw: Any,
    radixes: Sequence[int] | int | None = None,
    constant: bool = False,
) -> FrozenParameterGate:
    """Hypothesis strategy for generating `FrozenParameterGate`'s."""
    gate = draw(deferred(lambda: gates(radixes, False)))
    if gate.num_params == 0:
        return FrozenParameterGate(gate, {})
    max_idx = gate.num_params
    indices = integers(0, max_idx - 1)
    values = floats(allow_nan=False, allow_infinity=False, width=16)
    min_size = gate.num_params if constant else 0
    frozen_params = draw(
        dictionaries(
            indices,
            values,
            min_size=min_size,
            max_size=max_idx,
        ),
    )
    return FrozenParameterGate(gate, frozen_params)


@composite
def tagged_gates(
    draw: Any,
    radixes: Sequence[int] | int | None = None,
    constant: bool = False,
) -> TaggedGate:
    """Hypothesis strategy for generating `TaggedGate`'s."""
    gate = draw(deferred(lambda: gates(radixes, constant)))
    tag = draw(text())
    return TaggedGate(gate, tag)


# TODO: CircuitGate (circuit, move)
# TODO: ControlledGate (gate, num_controls, radixes)


@composite
def gates(
    draw: Any,
    radixes: Sequence[int] | int | None = None,
    constant: bool | None = None,
) -> Gate:
    """Hypothesis strategy for generating gates."""
    if radixes is None:
        radixes = 2

    if is_integer(radixes):
        radixes = [radixes] * draw(sampled_from([1, 2, 3, 4]))

    if not is_sequence_of_int(radixes):
        raise TypeError(
            'Expected sequence of integers or integer for radixes,'
            f'got {type(radixes)}.',
        )

    num_qudits = len(radixes)

    gate = draw(
        one_of(
            just(IdentityGate(num_qudits, radixes)),
            constant_unitary_gates(radixes),
            just(PauliGate(num_qudits)),
            just(VariableUnitaryGate(num_qudits, radixes)),
            sampled_from(gate_instances),
            dagger_gates(radixes, constant),
            tagged_gates(radixes, constant),
            frozen_gates(radixes, constant),
        ),
    )

    if constant is not None:
        assume(gate.is_constant() == constant)

    assume(sorted(gate.radixes) == sorted(radixes))
    assume(gate.num_params <= 128)

    return gate


@composite
def gates_and_params(
    draw: Any,
    radixes: Sequence[int] | int | None = None,
    constant: bool | None = None,
) -> tuple[Gate, RealVector]:
    """Hypothesis strategy for generating gates and parameters."""
    gate = draw(gates(radixes, constant))
    params = draw(
        lists(
            floats(allow_nan=False, allow_infinity=False, width=16),
            min_size=gate.num_params,
            max_size=gate.num_params,
        ),
    )
    return gate, params


@composite
def operations(
    draw: Any,
    radixes: Sequence[int] | int | None = None,
    constant: bool = False,
    max_qudit: int = 100,
) -> Operation:
    """Hypothesis strategy for generating operations."""
    gate = draw(gates(radixes, constant))
    location = draw(
        circuit_locations(
            max_qudit=max_qudit,
            min_size=gate.num_qudits,
            max_size=gate.num_qudits,
        ),
    )
    params = draw(
        one_of([
            lists(floats(), max_size=0),
            lists(
                floats(allow_nan=False, allow_infinity=False, width=16),
                min_size=gate.num_params,
                max_size=gate.num_params,
            ),
        ]),
    )
    return Operation(gate, location, params)


@composite
def circuits(
    draw: Any,
    radixes: Sequence[int] | int | None = None,
    max_gates: int = 10,
    exact_num_gates: bool = False,
    constant: bool = False,
) -> Circuit:
    """Hypothesis strategy for generating circuits."""

    if not isinstance(max_gates, int):
        raise TypeError(f'Expected int for max_gates, got: f{type(max_gates)}.')

    if not isinstance(constant, bool):
        raise TypeError(f'Expected bool for constant, got: {type(constant)}.')

    if radixes is None:
        radixes = 2

    if is_integer(radixes):
        radixes = [radixes] * draw(sampled_from([1, 2, 3, 4]))

    if not is_sequence_of_int(radixes):
        raise TypeError(
            'Expected sequence of integers or integer for radixes,'
            f'got {type(radixes)}.',
        )

    if exact_num_gates:
        num_gates = max_gates
    else:
        num_gates = draw(sampled_from(list(range(max_gates))))
    size = len(radixes)
    qubit_indices = list(range(size))
    idx_and_rdx = list(zip(qubit_indices, radixes))
    circuit = Circuit(size, radixes)

    for d in range(num_gates):
        gate_size = draw(sampled_from([1, 2, 3, 4]))
        gate_idx_and_rdx = draw(permutations(idx_and_rdx))[:gate_size]
        gate_location = list(zip(*gate_idx_and_rdx))[0]
        gate_radixes = list(zip(*gate_idx_and_rdx))[1]
        gate = draw(gates(gate_radixes, constant))
        params = floats(allow_nan=False, allow_infinity=False, width=16)
        num_params = gate.num_params
        gate_params = draw(
            lists(
                params,
                min_size=num_params,
                max_size=num_params,
            ),
        )
        circuit.append_gate(gate, gate_location, gate_params)

    return circuit


@composite
def circuit_locations(
    draw: Any,
    min_qudit: int = 0,
    max_qudit: int = 100,
    min_size: int = 1,
    max_size: int = 5,
) -> CircuitLocation:
    """Hypothesis strategy for generating `CircuitLocation` s."""
    idxs = integers(min_qudit, max_qudit)
    locations = lists(idxs, min_size=min_size, max_size=max_size, unique=True)
    return CircuitLocation(draw(locations))


@composite
def circuit_location_likes(
    draw: Any,
    min_qudit: int = 0,
    max_qudit: int = 100,
    min_size: int = 1,
    max_size: int = 5,
) -> CircuitLocationLike:
    """Hypothesis strategy for generating `CircuitLocationLike` s."""
    idxs = integers(min_qudit, max_qudit)
    locations = lists(idxs, min_size=min_size, max_size=max_size, unique=True)
    location = draw(locations)
    return draw(
        sampled_from(
            [location[0], location, CircuitLocation(location)],
        ),
    )


@composite
def circuit_points(
    draw: Any,
    max_cycle: int = 1000,
    max_qudit: int = 1000,
) -> CircuitPoint:
    """Hypothesis strategy for generating `CircuitPoint`'s."""
    cycle = draw(integers(0, max_cycle))
    qudit = draw(integers(0, max_qudit))
    return CircuitPoint(cycle, qudit)


@composite
def circuit_point_likes(
    draw: Any,
    max_cycle: int = 1000,
    max_qudit: int = 1000,
) -> CircuitPointLike:
    """Hypothesis strategy for generating `CircuitPointLike` objects."""
    point = draw(circuit_points(max_cycle, max_qudit))
    return draw(sampled_from([point, tuple(point)]))


@composite
def cycle_intervals(
    draw: Any,
    max_max_cycle: int = 1000,
    max_range: int = 100,
) -> CycleInterval:
    """Hypothesis strategy for generating `CycleInterval`'s."""
    lower = draw(integers(0, max_max_cycle))
    upper = draw(
        integers(lower, min(max_max_cycle, lower + max_range)),
    )
    return CycleInterval(lower, upper)


@composite
def circuit_regions(
    draw: Any,
    max_max_cycle: int = 1000,
    max_num_qudits: int = 20,
    max_volume: int = 200,
    max_qudit: int = 1000,
    empty: bool | None = None,
) -> CircuitRegion:
    """Hypothesis strategy for generating `CircuitRegion`'s."""
    if empty is True:
        return CircuitRegion({})

    region = draw(
        dictionaries(
            integers(0, max_qudit),
            cycle_intervals(max_max_cycle, max_volume // max_num_qudits),
            min_size=0 if empty is None else 1,
            max_size=max_num_qudits,
        ),
    )
    return CircuitRegion(region)


def everything_except(
    excluded_types: type | tuple[type, ...],
) -> SearchStrategy[Any]:
    """
    Hypothesis strategy to generate inputs of types not in `excluded_types`.

    References:
        https://hypothesis.readthedocs.io/en/latest/data.html
    """
    if not isinstance(excluded_types, tuple):
        excluded_types = tuple([excluded_types])

    checked_types = []
    for excluded_type in excluded_types:
        try:
            isinstance(0, excluded_type)
            checked_types.append(excluded_type)
        except TypeError:
            continue

    if len(checked_types) == 0:
        return from_type(type).flatmap(from_type)

    return (
        from_type(type)
        .flatmap(from_type)
        .filter(lambda x: not isinstance(x, tuple(checked_types)))
    )



"""This module implements the RecordStatsPass class."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class RecordStatsPass(BasePass):
    """
    The RecordStatsPass class.

    The RecordStatsPass stores stats about the circuit.
    """

    key = 'RecordStatsPass_stats_list'

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        stats: dict[str, Any] = {}
        stats['cycles'] = circuit.num_cycles
        stats['num_ops'] = circuit.num_operations
        stats['cgraph'] = circuit.coupling_graph
        stats['depth'] = circuit.depth
        stats['gate_counts'] = {
            gate: circuit.count(gate)
            for gate in circuit.gate_set
        }

        if self.key not in data:
            data[self.key] = []

        data[self.key].append(stats)

        _logger.info('Circuit Current Stats')
        _logger.info(f'{circuit.num_cycles} Cycles')
        _logger.info(f'{circuit.num_operations} Operation Count')
        _logger.info(f'{circuit.coupling_graph} Connectivity')
        for gate in circuit.gate_set:
            _logger.info(f'{circuit.count(gate)} {gate.name} Count')



"""This module implements the AStarHeuristic class."""
from __future__ import annotations

from bqskit.ir.circuit import Circuit
from bqskit.ir.opt.cost import CostFunctionGenerator
from bqskit.ir.opt.cost import HilbertSchmidtCostGenerator
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_real_number


class AStarHeuristic(HeuristicFunction):
    """
    The AStarHeuristic HeuristicFunction class.

    Defines a heuristic that combines the depth of the circuit with its cost.
    It generally gives similar quality results to DjikstraHeuristic, but with
    a drastic reduction in the number of instantiation calls.

    In pure search terms: f(p) = cost(p) + heuristic(p), this implements
    the f(p) component.
    """

    def __init__(
        self,
        heuristic_factor: float = 10.0,
        cost_factor: float = 1.0,
        cost_gen: CostFunctionGenerator = HilbertSchmidtCostGenerator(),
    ) -> None:
        """
        Construct a AStarHeuristic Function.

        Args:
            heuristic_factor (float): Scale the heuristic component by
                this value.

            cost_factor (float): Scale the cost component by this value.

            cost_gen (CostFunctionGenerator): This is used to generate
                cost functions used during evaluations.
        """
        if not is_real_number(heuristic_factor):
            raise TypeError(
                'Expected float for heuristic_factor, got %s.'
                % type(heuristic_factor),
            )

        if not is_real_number(cost_factor):
            raise TypeError(
                'Expected float for cost_factor, got %s.'
                % type(cost_factor),
            )

        if not isinstance(cost_gen, CostFunctionGenerator):
            raise TypeError(
                'Expected CostFunctionGenerator for cost_gen, got %s.'
                % type(cost_gen),
            )

        self.heuristic_factor = heuristic_factor
        self.cost_factor = cost_factor
        self.cost_gen = cost_gen

    def get_value(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> float:
        """Return the heuristic's value, see HeuristicFunction for more info."""
        cost = 0.0
        for gate in circuit.gate_set:
            if gate.num_qudits == 1:
                continue
            cost += float(circuit.count(gate))

        heuristic = self.cost_gen.calc_cost(circuit, target)
        return self.heuristic_factor * heuristic + self.cost_factor * cost



"""This module implements the ArbitraryCPhaseGate."""
from __future__ import annotations

from typing import Sequence

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass
from bqskit.utils.typing import is_sequence


class ArbitraryCPhaseGate(
    Gate,
    DifferentiableUnitary,
    CachedClass,
    LocallyOptimizableUnitary,
):
    """A gate representing an arbitrary qudit controlled phase rotation."""

    _num_params = 1

    def __init__(self, radixes: Sequence[int] = []) -> None:
        if len(radixes) == 0:
            radixes = [2, 2]

        if not is_sequence(radixes):
            raise TypeError(
                f'Expected sequence for radixes, got {type(radixes)}.',
            )

        if any(r <= 1 for r in radixes):
            raise TypeError('Invalid radixes, all radixes must be >= 2.')

        self._num_qudits = len(radixes)
        self._radixes = tuple(radixes)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)
        U = np.identity(self.dim, dtype=np.complex128)
        U[-1, -1] = np.exp(1j * params[0])
        return UnitaryMatrix(U)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)
        dU = np.zeros((1, self.dim, self.dim), dtype=np.complex128)
        dU[-1, -1, -1] = 1j * np.exp(1j * params[0])
        return dU

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        p = env_matrix[-1, -1]
        return [-np.arctan2(p.imag, p.real)]

    def __eq__(self, other: object) -> bool:
        """Check if `self` equals `other`."""
        if not isinstance(other, ArbitraryCPhaseGate):
            return False
        return self.radixes == other.radixes

    def __hash__(self) -> int:
        return hash(self.radixes)



"""This module implements the ExtractDiagonalPass."""
from __future__ import annotations

from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import DiagonalGate
from bqskit.ir.gates import VariableUnitaryGate
from bqskit.ir.gates.constant import CNOTGate
from bqskit.ir.operation import Operation
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


theorized_bounds = [0, 0, 3, 14, 61, 252]


def construct_linear_ansatz(num_qudits: int) -> Circuit:
    """
    Generate a linear ansatz for extracting the diagonal of a unitary.

    This ansatz consists of a `num_qudits` width Diagonal Gate followed
    by a ladder of CNOTs and single qubit gates. Right now, we try to use
    one fewer CNOT than the theorized minimum number of CNOTs to represent
    the unitary.

    This ansatz is simply a heuristic and does not have theoretical
    backing. However, we see that for unitaries up to 5 qubits, this
    ansatz does succeed most of the time with a threshold of 1e-8.
    """
    theorized_num = theorized_bounds[num_qudits]
    circuit = Circuit(num_qudits)
    circuit.append_gate(DiagonalGate(num_qudits), tuple(range(num_qudits)))
    for i in range(num_qudits):
        circuit.append_gate(VariableUnitaryGate(1), (i,))
    for _ in range(theorized_num // (num_qudits - 1)):
        # Apply n - 1 linear CNOTs
        for i in range(num_qudits - 1):
            circuit.append_gate(CNOTGate(), (i, i + 1))
            circuit.append_gate(VariableUnitaryGate(1), (i,))
            circuit.append_gate(VariableUnitaryGate(1), (i + 1,))
    return circuit


class ExtractDiagonalPass(BasePass):
    """
    A pass that attempts to extract a diagonal matrix from a unitary matrix.

    https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=1269020

    While there is a known algorithm for 2-qubit gates, we utilize
    synthesis methods instead to scale to wider qubit gates.

    As a heuristic, we attempt to extrac the diagonal using a linear chain
    ansatz of CNOT gates. We have found that up to 5 qubits, this ansatz
    does succeed for most unitaries with fewer CNOTs than the theoretical
    minimum number of CNOTs (utilizing the power of the Diagonal Gate in front).
    """

    def __init__(
        self,
        qudit_size: int = 2,
        success_threshold: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        # We only support diagonal extraction for 2-5 qubits
        assert qudit_size >= 2 and qudit_size <= 5
        self.qudit_size = qudit_size
        self.success_threshold = success_threshold
        self.cost = cost
        self.instantiate_options: dict[str, Any] = {
            'cost_fn_gen': self.cost,
            'min_iters': 0,
            'diff_tol_r': 1e-4,
            'multistarts': 16,
            'method': 'qfactor',
        }
        self.instantiate_options.update(instantiate_options)
        super().__init__()

    async def decompose(
        self,
        op: Operation,
        target: UnitaryMatrix,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        success_threshold: float = 1e-14,
        instantiate_options: dict[str, Any] = {},
    ) -> tuple[
        Operation | None,
        Circuit,
    ]:
        """
        Return the circuit that is generated from one levl of QSD.

        Args:
            op (Operation): The VariableUnitaryGate Operation to decompose.
            target (UnitaryMatrix): The target unitary.
            cost (CostFunctionGenerator): The cost function generator to
            determine if we have succeeded in decomposing the gate.
            success_threshold (float): The threshold for the cost function.
            instantiate_options (dict[str, Any]): The options to pass to the
            instantiate method.
        """

        circ = Circuit(op.gate.num_qudits)

        if op.gate.num_qudits == 2:
            # For now just try for 2 qubit
            circ.append_gate(DiagonalGate(op.gate.num_qudits), (0, 1))
            circ.append_gate(VariableUnitaryGate(op.gate.num_qudits - 1), (0,))
            circ.append_gate(VariableUnitaryGate(op.gate.num_qudits - 1), (1,))
            circ.append_gate(CNOTGate(), (0, 1))
            circ.append_gate(VariableUnitaryGate(op.gate.num_qudits - 1), (0,))
            circ.append_gate(VariableUnitaryGate(op.gate.num_qudits - 1), (1,))
            circ.append_gate(CNOTGate(), (0, 1))
            circ.append_gate(VariableUnitaryGate(op.gate.num_qudits - 1), (0,))
            circ.append_gate(VariableUnitaryGate(op.gate.num_qudits - 1), (1,))
        elif op.gate.num_qudits == 3:
            circ = construct_linear_ansatz(op.gate.num_qudits)
        else:
            circ = construct_linear_ansatz(op.gate.num_qudits)

        instantiated_circ = circ.instantiate(
            target=target,
            **instantiate_options,
        )

        if cost.calc_cost(instantiated_circ, target) < success_threshold:
            diag_op = instantiated_circ.pop((0, 0))
            return diag_op, instantiated_circ

        default_circ = Circuit(op.gate.num_qudits)
        default_circ.append_gate(
            op.gate,
            tuple(range(op.gate.num_qudits)), op.params,
        )
        return None, default_circ

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Synthesize `utry`, see :class:`SynthesisPass` for more."""
        num_ops = 0

        num_gates_to_consider = circuit.count(
            VariableUnitaryGate(self.qudit_size),
        )

        while num_gates_to_consider > 1:
            # Find last Unitary
            all_ops = list(circuit.operations_with_cycles(reverse=True))
            found = False
            for cyc, op in all_ops:
                if (
                    isinstance(op.gate, VariableUnitaryGate)
                    and op.gate.num_qudits in [2, 3, 4]
                ):
                    if found:
                        merge_op = op
                        merge_pt = (cyc, op.location[0])
                        merge_location = op.location
                        break
                    else:
                        num_ops += 1
                        gate = op
                        pt = (cyc, op.location[0])
                        found = True
            diag_op, circ = await self.decompose(
                gate,
                cost=self.cost,
                target=gate.get_unitary(),
                success_threshold=self.success_threshold,
                instantiate_options=self.instantiate_options,
            )

            circuit.replace_with_circuit(pt, circ, as_circuit_gate=True)
            num_gates_to_consider -= 1
            # Commute Diagonal into next op
            if diag_op:
                new_mat = diag_op.get_unitary() @ merge_op.get_unitary()
                circuit.replace_gate(
                    merge_pt, merge_op.gate, merge_location,
                    VariableUnitaryGate.get_params(new_mat),
                )

        circuit.unfold_all()



"""This module implements the RYGate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class RYGate(
    QubitGate,
    DifferentiableUnitary,
    LocallyOptimizableUnitary,
    CachedClass,
):
    """
    A gate representing an arbitrary rotation around the Y axis.

    It is given by the following parameterized unitary:

    .. math::

        \\begin{pmatrix}
        \\cos{\\frac{\\theta}{2}} & -\\sin{\\frac{\\theta}{2}} \\\\
        \\sin{\\frac{\\theta}{2}} & \\cos{\\frac{\\theta}{2}} \\\\
        \\end{pmatrix}
    """

    _num_qudits = 1
    _num_params = 1
    _qasm_name = 'ry'

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        cos = np.cos(params[0] / 2)
        sin = np.sin(params[0] / 2)

        return UnitaryMatrix(
            [
                [cos, -sin],
                [sin, cos],
            ],
        )

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        dcos = -np.sin(params[0] / 2) / 2
        dsin = np.cos(params[0] / 2) / 2

        return np.array(
            [
                [
                    [dcos, -dsin],
                    [dsin, dcos],
                ],
            ], dtype=np.complex128,
        )

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        self.check_env_matrix(env_matrix)
        a = np.real(env_matrix[0, 0] + env_matrix[1, 1])
        b = np.real(env_matrix[1, 0] - env_matrix[0, 1])
        theta = 2 * np.arccos(a / np.sqrt(a ** 2 + b ** 2))
        theta *= -1 if b > 0 else 1
        return [theta]



"""This module implements the AndPredicate class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


class AndPredicate(PassPredicate):
    """
    The AndPredicate class.

    The AndPredicate takes two predicates and returns true iff both are true.
    """

    def __init__(self, p1: PassPredicate, p2: PassPredicate) -> None:
        """
        Construct a AndPredicate.

        Args:
            p1 (PassPredicate): The first predicate.

            p2 (PassPredicate): The second predicate.
        """

        if not isinstance(p1, PassPredicate):
            raise TypeError(f'Expected PassPredicate, got {type(p1)}.')

        if not isinstance(p2, PassPredicate):
            raise TypeError(f'Expected PassPredicate, got {type(p2)}.')

        self.p1 = p1
        self.p2 = p2

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return self.p1(circuit, data) and self.p2(circuit, data)



"""This module implements the SubtopologySelectionPass pass."""
from __future__ import annotations

import copy
import itertools as it
from typing import Any
from typing import Iterable
from typing import Sequence

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.machine import MachineModel
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.qis.graph import CouplingGraph
from bqskit.utils.typing import is_integer


def powerset(iterable: Iterable[Any]) -> Iterable[Any]:
    """Powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    all_combos = (it.combinations(s, r) for r in range(len(s) + 1))
    return it.chain.from_iterable(all_combos)


def all_coupling_graphs_of_size(n: int) -> list[CouplingGraph]:
    """Calculates all valid coupling graphs with `n` qudits."""
    if not is_integer(n):
        raise TypeError(f'Expected integer for n, got {type(n)}.')

    if n <= 1:
        raise ValueError(f'Smallest valid graph is 2, got {n}.')

    graphs: list[list[list[tuple[int, int]]]] = [[[], [(0, 1)]]]

    while len(graphs) + 1 < n:
        i = len(graphs) + 1
        extended_graph_list = []
        for graph in graphs[-1]:
            for nodes in powerset(range(i)):
                edges = copy.deepcopy(graph)

                for node in nodes:
                    edges.append((node, i))

                extended_graph_list.append(edges)
        graphs.append(extended_graph_list)

    to_ret = []
    for g in graphs[-1]:
        cg = CouplingGraph(g, n)
        if cg.is_fully_connected():
            to_ret.append(cg)

    return to_ret


class GraphDAGNode():
    """A node in the GraphDAG."""

    def __init__(self, graph_id: int):
        self.indices: list[int] = [graph_id]
        self.predecessors: list[GraphDAGNode] = []
        self.successors: list[GraphDAGNode] = []

    def peek_index(self) -> int:
        """Provide index of this kind of graph in list."""
        return self.indices[0]


class GraphDAG():
    """
    DAG for organizing the embeddability relations of graphs.

    The DAG is organized so that a node is guaranteed to be embedded within all
    of its successor nodes.
    """

    def __init__(self, graph_list: Sequence[CouplingGraph]):
        self.roots: list[GraphDAGNode] = []
        self.leafs: list[GraphDAGNode] = []
        self.graph_list = list(graph_list)
        if not self.graph_list[0]._edges == set():
            self.graph_list.insert(0, CouplingGraph([]))
        self._create_DAG()
        self._insert_node(0)

    def _create_DAG(self) -> None:
        for graph_id in range(len(self.graph_list)):
            self._insert_node(graph_id)

    def _forward_path_exists(self, node_a: Any, node_b: Any) -> bool:
        """Return True if a forward path exists from node_a to node_b."""
        if node_a is node_b or node_b in node_a.successors:
            return True
        to_check = node_a.successors.copy()
        for node in to_check:
            if node_b in node.successors:
                return True
            else:
                to_check.extend(node.successors)
        return False

    def _insert_node(self, graph_id: int) -> None:
        """
        Insert a new node corresponding to the graph `graph_list[graph_id]` into
        the DAG.

        Append it to the "least embeddable" graph that is still embedded within
        it.
        """
        node = GraphDAGNode(graph_id)
        if len(self.roots) == 0:
            self.roots.append(node)
            self.leafs.append(node)
            return

        graph = self.graph_list[graph_id]
        nodes_to_check = self.leafs.copy()

        # Check for node to be predessor
        for other_node in nodes_to_check:
            other_graph = self.graph_list[
                other_node.peek_index()
            ]

            if other_graph.is_embedded_in(graph):
                # Handle isomorphic case
                if graph.is_embedded_in(other_graph):
                    other_node.indices.append(graph_id)
                    return
                if not self._forward_path_exists(other_node, node):
                    other_node.successors.append(node)
                    node.predecessors.append(other_node)

                # Check if we were at a leaf
                if other_node in self.leafs:
                    self.leafs.remove(other_node)
                if node not in self.leafs:
                    self.leafs.append(node)
            else:
                # Only check predecessors if node is not embedded in
                # leaf and there is not already a direct forward path
                # to the node
                nodes_to_check += other_node.predecessors

    def get_embedded_indices(
        self,
        graphs: list[CouplingGraph] | CouplingGraph,
    ) -> list[int]:
        """Given a list of graphs, return the indices of subgraphs in the
        `self.graph_list` list that are embedded in a subgraph in `graphs`."""
        embedded_indices: set[int] = set()
        if isinstance(graphs, CouplingGraph):
            graphs = [graphs]

        for graph in graphs:
            # Check least embeddable graphs first
            to_check: list[GraphDAGNode] = self.leafs.copy()
            for node in to_check:
                subgraph = self.graph_list[node.peek_index()]
                if not subgraph.is_embedded_in(graph):
                    # If not embedded, check predecessors
                    to_check.extend(node.predecessors)
                else:
                    # If embedded, add node's indices and all predecessors
                    new_id = node.peek_index()
                    if new_id not in embedded_indices:
                        embedded_indices = embedded_indices.union(
                            set(node.indices),
                        )
                        preds = node.predecessors
                        for p in preds:
                            if p.peek_index() not in embedded_indices:
                                embedded_indices = embedded_indices.union(
                                    set(p.indices),
                                )
                                preds.extend(p.predecessors)
        return sorted(list(embedded_indices))


def filter_compatible_subgraphs(
    candidate_subgraphs: Sequence[CouplingGraph],
    machine: MachineModel,
    blocksize: int | None = None,
) -> list[CouplingGraph]:
    """
    Filter the candidate subgraphs, returning the ones appearing in `machine`.

    Given a list of candidate synthesis subgraphs and a MachineModel with a
    valid defined CouplingGraph, return a list of candidate subgraphs that are
    embedded within the machine.

    Arguments:
        candidate_subgraphs (list[CouplingGraph]): A list containing subgraphs
            that should be checked for embeddability within some MachineModel.

        machine (MachineModel): A MachineModel with a valid and specified
            CouplingGraph.

        blocksize (int | None): Maximum desired subgraph size. If unspecified,
            the entire CouplingGraph will be searched. It is highly recommended
            that the blocksize be provided for performance reasons.
            (Default: None)

    Returns:
        (list[CouplingGraph]): A list of coupling graphs that are both in the
            `candidate_subgraphs` list and the `machine`'s CouplingGraph.

    Raises:
        ValueError: If the CouplingGraph of the `machine` variable is either
            invalid or unspecified.
    """
    graph = machine.coupling_graph
    if blocksize is None:
        blocksize = graph.num_qudits

    # Sort graphs by number of edge
    locations = graph.get_subgraphs_of_size(blocksize)
    induced_subgraphs: list[CouplingGraph] = sorted(
        (
            graph.get_subgraph(l)
            for l in locations
        ),
        key=lambda x: -len(x),
    )
    candidates = sorted(candidate_subgraphs, key=lambda x: len(x))
    dag = GraphDAG(candidates)
    indices_of_interest = dag.get_embedded_indices(induced_subgraphs)

    to_return = [
        candidates[i] for i in range(len(candidates))
        if i in indices_of_interest
    ]
    # Do not return empty set used as root
    if to_return[0]._edges == set():
        return to_return[1:]
    return to_return


class SubtopologySelectionPass(BasePass):
    """Pass that selects necessary subtopologies from the model."""

    key = ForEachBlockPass.pass_down_key_prefix + 'sub_topologies'

    def __init__(self, block_size: int) -> None:
        """
        Construct a SubtopologySelectionPass.

        Args:
            block_size (int): The max block size to select subtopologies for.

        Raises:
            ValueError: If block_size is <= 1.
        """
        if not is_integer(block_size):
            raise TypeError(f'Expected integer, got {type(block_size)}.')

        if block_size <= 1:
            raise ValueError(f'Expected integer > 1, got {block_size}.')

        self.block_size = block_size

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        model = data.model
        n = model.num_qudits
        cg = model.coupling_graph
        all_to_all = len(cg._edges) == n * (n - 1) / 2

        tops = {}
        for i in range(2, self.block_size + 1):
            if all_to_all:
                tops[i] = [CouplingGraph.all_to_all(i)]
            else:
                all_tops = all_coupling_graphs_of_size(i)
                tops[i] = filter_compatible_subgraphs(all_tops, model, i)

        data[self.key] = tops



"""This module implements the CNOTToCZPass."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.gates.constant.cz import CZGate
from bqskit.ir.gates.constant.h import HGate
from bqskit.ir.operation import Operation


class CNOTToCZPass(BasePass):
    """
    The CNOTToCZPass class.

    This uses a rule to convert CNOTs to CZs.
    """

    def __init__(self) -> None:
        """Construct a CNOTToCZPass."""
        circuit = Circuit(2)
        circuit.append_gate(HGate(), 1)
        circuit.append_gate(CZGate(), (0, 1))
        circuit.append_gate(HGate(), 1)
        self.cg = CircuitGate(circuit)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Find all cnots
        cnot_points = []
        for cycle, op in circuit.operations_with_cycles():
            if isinstance(op.gate, CNOTGate):
                cnot_points.append((cycle, op.location[0]))

        # Create new operations
        ops = [
            Operation(
                self.cg,
                circuit[p].location,
                self.cg._circuit.params,
            )
            for p in cnot_points
        ]

        # Replace cnots with new ops
        circuit.batch_replace(cnot_points, ops)
        circuit.unfold_all()  # TODO: Replace with batch_unfold



"""This module implements the RandomStartGenerator class."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from bqskit.ir.opt.multistartgen import MultiStartGenerator
from bqskit.utils.typing import is_integer

if TYPE_CHECKING:
    import numpy.typing as npt
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.state.state import StateVector
    from bqskit.qis.state.system import StateSystem
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class RandomStartGenerator(MultiStartGenerator):
    """A start generator that selects random points uniformily across [0, 2π)"""

    def gen_starting_points(
        self,
        multistarts: int,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> list[npt.NDArray[np.float64]]:
        """
        Generate `multistarts` starting points for instantiation.

        Args:
            multistarts (int): The number of starting points to generate.

            circuit (Circuit): The circuit to generate the points for.

            target (UnitaryMatrix | StateVector): The target.

        Return:
            (list[npt.NDArray[np.float64]]): List of starting inputs for
                instantiation.

        Raises:
            ValueError: If `multistarts` is not a positive integer.
        """

        if not is_integer(multistarts):
            raise TypeError(
                'Expected int for multistarts, got %s.' % type(multistarts),
            )

        if multistarts <= 0:
            raise ValueError(
                'Expected positive integer for multistarts'
                ', got %d' % multistarts,
            )

        return [
            2 * np.pi * np.random.random(circuit.num_params)
            for i in range(multistarts)
        ]



"""This module implements the DistributedPredicate class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class DistributedPredicate(PassPredicate):
    """
    The DistributedPredicate class.

    The DistributedPredicate returns true if the targeted machine is distributed
    across multiple chips.
    """

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return data.model.coupling_graph.is_distributed()



"""The Berkeley Quantum Synthesis Toolkit Python Package."""
from __future__ import annotations

from typing import Any

from bqskit._logging import disable_logging
from bqskit._logging import enable_logging
from bqskit._version import __version__  # noqa: F401
from bqskit._version import __version_info__  # noqa: F401


def __getattr__(name: str) -> Any:
    # Lazy imports
    if name == 'compile':
        from bqskit.compiler.compile import compile
        return compile

    if name == 'Circuit':
        from bqskit.ir.circuit import Circuit
        return Circuit

    if name == 'MachineModel':
        from bqskit.compiler.machine import MachineModel
        return MachineModel

    raise AttributeError(f'module {__name__} has no attribute {name}')


__all__ = [
    'compile',
    'MachineModel',
    'Circuit',
    'enable_logging',
    'disable_logging',
]



"""This module implements the U8Gate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.qutritgate import QutritGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class U8Gate(QutritGate, DifferentiableUnitary, CachedClass, GeneralGate):
    """The U8 single qutrit gate."""

    _num_qudits = 1
    _num_params = 8

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        s1 = np.sin(params[0])
        c1 = np.cos(params[0])
        s2 = np.sin(params[1])
        c2 = np.cos(params[1])
        s3 = np.sin(params[2])
        c3 = np.cos(params[2])

        p1 = np.exp(1j * params[3])
        m1 = np.exp(-1j * params[3])
        p2 = np.exp(1j * params[4])
        m2 = np.exp(-1j * params[4])
        p3 = np.exp(1j * params[5])
        m3 = np.exp(-1j * params[5])
        p4 = np.exp(1j * params[6])
        m4 = np.exp(-1j * params[6])
        p5 = np.exp(1j * params[7])
        m5 = np.exp(-1j * params[7])

        return UnitaryMatrix(
            [
                [
                    c1 * c2 * p1,
                    s1 * p3,
                    c1 * s2 * p4,
                ],
                [
                    s2 * s3 * m4 * m5 - s1 * c2 * c3 * p1 * p2 * m3,
                    c1 * c3 * p2,
                    -c2 * s3 * m1 * m5 - s1 * s2 * c3 * p2 * m3 * p4,
                ],
                [
                    -s1 * c2 * s3 * p1 * m3 * p5 - s2 * c3 * m2 * m4,
                    c1 * s3 * p5,
                    c2 * c3 * m1 * m2 - s1 * s2 * s3 * m3 * p4 * p5,
                ],
            ],
        )

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        s1 = np.sin(params[0])
        c1 = np.cos(params[0])
        s2 = np.sin(params[1])
        c2 = np.cos(params[1])
        s3 = np.sin(params[2])
        c3 = np.cos(params[2])

        p1 = np.exp(1j * params[3])
        m1 = np.exp(-1j * params[3])
        p2 = np.exp(1j * params[4])
        m2 = np.exp(-1j * params[4])
        p3 = np.exp(1j * params[5])
        m3 = np.exp(-1j * params[5])
        p4 = np.exp(1j * params[6])
        m4 = np.exp(-1j * params[6])
        p5 = np.exp(1j * params[7])
        m5 = np.exp(-1j * params[7])

        return np.array(
            [
                [  # wrt params[0]
                    [-s1 * c2 * p1, c1 * p3, -s1 * s2 * p4],
                    [
                        -c1 * c2 * c3 * p1 * p2 * m3, -s1 * c3 * p2,
                        -c1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        -c1 * c2 * s3 * p1 * m3 * p5, -s1 * s3 * p5,
                        -c1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[1]
                    [-c1 * s2 * p1, 0, c1 * c2 * p4],
                    [
                        c2 * s3 * m4 * m5 + s1 * s2 * c3 * p1 * p2 * m3, 0,
                        s2 * s3 * m1 * m5 - s1 * c2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        s1 * s2 * s3 * p1 * m3 * p5 - c2 * c3 * m2 * m4, 0,
                        -s2 * c3 * m1 * m2 - s1 * c2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[2]
                    [0, 0, 0],
                    [
                        s2 * c3 * m4 * m5 + s1 * c2 * s3 * p1 * p2 * m3,
                        -c1 * s3 * p2,
                        -c2 * c3 * m1 * m5 + s1 * s2 * s3 * p2 * m3 * p4,
                    ],
                    [
                        -s1 * c2 * c3 * p1 * m3 * p5 + s2 * s3 * m2 * m4,
                        c1 * c3 * p5,
                        -c2 * s3 * m1 * m2 - s1 * s2 * c3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[3]
                    [1j * c1 * c2 * p1, 0, 0],
                    [
                        -1j * s1 * c2 * c3 * p1 * p2 * m3,
                        0,
                        1j * c2 * s3 * m1 * m5,
                    ],
                    [
                        -1j * s1 * c2 * s3 * p1 * m3 * p5,
                        0, -1j * c2 * c3 * m1 * m2,
                    ],
                ],

                [  # wrt params[4]
                    [0, 0, 0],
                    [
                        -1j * s1 * c2 * c3 * p1 * p2 * m3, 1j * c1 * c3 * p2,
                        -1j * s1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [1j * s2 * c3 * m2 * m4, 0, -1j * c2 * c3 * m1 * m2],
                ],

                [  # wrt params[5]
                    [0, 1j * s1 * p3, 0],
                    [
                        1j * s1 * c2 * c3 * p1 * p2 * m3, 0,
                        1j * s1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        1j * s1 * c2 * s3 * p1 * m3 * p5, 0,
                        1j * s1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[6]
                    [0, 0, 1j * c1 * s2 * p4],
                    [
                        -1j * s2 * s3 * m4 * m5,
                        0,
                        -1j * s1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        1j * s2 * c3 * m2 * m4,
                        0,
                        -1j * s1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[7]
                    [0, 0, 0],
                    [-1j * s2 * s3 * m4 * m5, 0, 1j * c2 * s3 * m1 * m5],
                    [
                        -1j * s1 * c2 * s3 * p1 * m3 * p5, 1j * c1 * s3 * p5,
                        -1j * s1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],
            ],
        )

    def calc_params(self, utry: UnitaryMatrix) -> list[float]:
        """Return the parameters for this gate to implement `utry`"""
        if utry.radixes != (3,):
            raise ValueError('Expected single-qutrit unitary.')

        params = [0.0] * 8

        mag = np.linalg.det(utry.numpy) ** (-1 / 3)
        su = mag * utry

        params[5] = np.arctan2(su[0][1].imag, su[0][1].real)

        s1 = (su[0][1] * np.exp(-1j * params[5])).real
        params[0] = np.arcsin(s1)

        c2p1 = su[0][0] / np.cos(params[0])
        params[3] = np.arctan2(c2p1.imag, c2p1.real)

        c2 = (c2p1 * np.exp(-1j * params[3])).real
        params[1] = np.arccos(c2)

        c3p2 = su[1][1] / np.cos(params[0])
        params[4] = np.arctan2(c3p2.imag, c3p2.real)

        c3 = (c3p2 * np.exp(-1j * params[4])).real
        params[2] = np.arccos(c3)

        p4 = su[0][2] / (np.cos(params[0]) * np.sin(params[1]))
        params[6] = np.arctan2(p4.imag, p4.real)

        p5 = su[2][1] / (np.cos(params[0]) * np.sin(params[2]))
        params[7] = np.arctan2(p5.imag, p5.real)

        return params



"""This module implements the PAMVerificationSequence and helper passes."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import PermutationGate
from bqskit.ir.gates import TaggedGate
from bqskit.ir.gates.constant.unitary import ConstantUnitaryGate
from bqskit.passes.alias import PassAlias
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.passes.mapping.pam import PAMBlockResultDict
from bqskit.passes.mapping.routing.pam import PAMRoutingPass
from bqskit.passes.partitioning import QuickPartitioner
from bqskit.passes.util.unfold import UnfoldPass
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import Sequence


class TagPAMBlockDataPass(BasePass):
    """Tag the blocks with the PAM block data."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        if PAMRoutingPass.out_data_key not in data:
            raise RuntimeError('PAMRoutingPass must be run to verify results.')

        block_datas: PAMBlockResultDict = data[PAMRoutingPass.out_data_key]
        for block_point, block_data in block_datas.items():
            op = circuit[block_point]
            tagged_gate = TaggedGate(op.gate, block_data)
            circuit.replace_gate(
                block_point,
                tagged_gate,
                op.location,
                op.params,
            )

        del data[PAMRoutingPass.out_data_key]


class CalculatePAMErrorsPass(BasePass):
    """Calculates error of a panel consisting of blocks tagged with PAM data."""

    @staticmethod
    def get_opp_perm(in_perm: Sequence[int]) -> Sequence[int]:
        return tuple(in_perm.index(i) for i in range(len(in_perm)))

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        # calculate approximate (current) panel unitary
        current_unitary = circuit.get_unitary()

        # calculate exact panel unitary
        exact_circuit = Circuit(circuit.num_qudits, circuit.radixes)
        for op in circuit:
            if isinstance(op.gate, TaggedGate):
                pi = op.gate.tag['pre_perm']
                pf = op.gate.tag['post_perm']
                in_utry = op.gate.tag['original_utry']
                n = in_utry.num_qudits
                exact_circuit.append_gate(
                    PermutationGate(
                        n, CalculatePAMErrorsPass.get_opp_perm(pi),
                    ), op.location,
                )
                exact_circuit.append_gate(
                    ConstantUnitaryGate(in_utry), op.location,
                )
                exact_circuit.append_gate(
                    PermutationGate(
                        n, CalculatePAMErrorsPass.get_opp_perm(pf),
                    ), op.location,
                )

            else:
                exact_circuit.append_gate(op.gate, op.location, op.params)

        exact_unitary = exact_circuit.get_unitary()

        # calculate error
        data.update_error_mul(current_unitary.get_distance_from(exact_unitary))


class UnTagPAMBlockDataPass(BasePass):
    """Untag the blocks with the PAM block data."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        for cycle, op in circuit.operations_with_cycles():
            if isinstance(op.gate, TaggedGate):
                circuit.replace_gate(
                    (cycle, op.location[0]),
                    op.gate.gate,
                    op.location,
                    op.params,
                )


class PAMVerificationSequence(PassAlias):
    """Calculates the error of a PAM sequence."""

    def __init__(self, error_sim_size: int = 8) -> None:
        """
        Construct a PAMVerificationPass.

        Args:
            error_sim_size (int): The block size to use during error
                calculations.
        """
        if not is_integer(error_sim_size):
            raise TypeError(f'Expected integer, got {type(error_sim_size)}.')

        if error_sim_size < 2:
            raise ValueError('Expected positive integer greater than 1.')

        self.error_sim_size = error_sim_size

    def get_passes(self) -> list[BasePass]:
        """Return the aliased passes, see :class:`PassAlias` for more info."""
        return [
            TagPAMBlockDataPass(),
            QuickPartitioner(self.error_sim_size),
            ForEachBlockPass(CalculatePAMErrorsPass()),
            UnfoldPass(),
            UnTagPAMBlockDataPass(),
        ]



"""This module contains the logging configuration and methods for BQSKit."""
from __future__ import annotations

import logging
from sys import stdout as _stdout


_logging_initialized = False


def enable_logging(verbose: bool = False) -> None:
    """
    Enable logging for BQSKit.

    Args:
        verbose (bool): If set to True, will print more verbose messages.
            Defaults to False.
    """
    global _logging_initialized
    if not _logging_initialized:
        _logger = logging.getLogger('bqskit')
        _handler = logging.StreamHandler(_stdout)
        _handler.setLevel(0)
        _fmt_header = '%(asctime)s.%(msecs)03d - %(levelname)-8s |'
        _fmt_message = ' %(name)s: %(message)s'
        _fmt = _fmt_header + _fmt_message
        _formatter = logging.Formatter(_fmt, '%H:%M:%S')
        _handler.setFormatter(_formatter)
        _logger.addHandler(_handler)
        _logging_initialized = True

    level = logging.DEBUG if verbose else logging.INFO
    logging.getLogger('bqskit').setLevel(level)


def disable_logging() -> None:
    """Disable logging for BQSKit."""
    logging.getLogger('bqskit').setLevel(logging.CRITICAL)



"""This module implements functions for working with IBM Backends."""
from __future__ import annotations

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from qiskit.providers import BackendV1

from bqskit.compiler.machine import MachineModel
from bqskit.ir.gate import Gate
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.gates.constant.cz import CZGate
from bqskit.ir.gates.constant.sx import SXGate
from bqskit.ir.gates.constant.x import XGate
from bqskit.ir.gates.constant import IdentityGate
from bqskit.ir.gates.constant import ECRGate
from bqskit.ir.gates.parameterized import RZGate
from bqskit.ir.gates.parameterized.u1 import U1Gate
from bqskit.ir.gates.parameterized.u2 import U2Gate
from bqskit.ir.gates.parameterized.u3 import U3Gate


def model_from_backend(backend: BackendV1) -> MachineModel:
    """Create a machine model for a IBM Backend."""
    config = backend.configuration()
    num_qudits = config.n_qubits
    gate_set = _basis_gate_str_to_bqskit_gate(config.basis_gates)
    coupling_map = list({tuple(sorted(e)) for e in config.coupling_map})
    return MachineModel(num_qudits, coupling_map, gate_set)


def _basis_gate_str_to_bqskit_gate(basis_gates: list[str]) -> set[Gate]:
    gate_set: set[Gate] = set()
    if len(basis_gates) > 10:
        return {CNOTGate(), RZGate(), SXGate()}
    for basis_gate in basis_gates:
        if basis_gate == 'cx':
            gate_set.add(CNOTGate())
        elif basis_gate == 'cz':
            gate_set.add(CZGate())
        elif basis_gate == 'ecr':
            gate_set.add(ECRGate())
        elif basis_gate == 'u3':
            gate_set.add(U3Gate())
        elif basis_gate == 'u2':
            gate_set.add(U2Gate())
        elif basis_gate == 'u1':
            gate_set.add(U1Gate())
        elif basis_gate == 'rz':
            gate_set.add(RZGate())
        elif basis_gate == 'x':
            gate_set.add(XGate())
        elif basis_gate == 'sx':
            gate_set.add(SXGate())
        elif basis_gate == 'p':
            gate_set.add(RZGate())
        elif basis_gate == 'id':
            gate_set.add(IdentityGate())
    return gate_set



"""This module implements the ShiftGate."""
from __future__ import annotations

import numpy as np

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class ShiftGate(ConstantGate, QuditGate):
    """
    The one-qudit shift (X) gate. This is a Weyl-Heisenberg gate.

    This gate shifts the state of a qudit up by one level modulo. For
    example, the shift gate on a qubit is the Pauli-X gate. The shift
    gate on a qutrit is the following matrix:

    .. math::

        \\begin{pmatrix}
        0 & 0 & 1 \\\\
        1 & 0 & 0 \\\\
        0 & 1 & 0 \\\\
        \\end{pmatrix}


    The shift gate is generally given by the following formula:

    .. math::
        \\begin{equation}
            X = \\sum_a |a + 1 mod d ><a|
        \\end{equation}

    where d is the number of levels (2 levels is a qubit,
    3 levels is a qutrit, etc.)

    References:
        - https://arxiv.org/pdf/2302.07966.pdf
    """

    _num_qudits = 1

    def __init__(self, radix: int = 2) -> None:
        """
        Construct a ShiftGate.

        Args:
            radix (int): The number of qudit levels (>=2). Defaults to
                qutrits.

        Raises:
            ValueError: if radix < 2
        """

        if not is_integer(radix):
            raise TypeError(f'Expected integer for radix, got {type(radix)}.')

        if radix < 2:
            raise ValueError(f'Radix must be greater than 1, got {radix}.')

        self._radix = radix

        # Calculate unitary
        matrix = np.zeros([radix, radix])
        for j in range(radix):
            matrix[(j + 1) % radix, j] = 1
        self._utry = UnitaryMatrix(matrix, self.radixes)



"""
=======================================================
Quantum Information Science Library (:mod:`bqskit.qis`)
=======================================================

.. currentmodule:: bqskit.qis

The `bqskit.qis` package contains class definitions
for common objects found in quantum information science.

The most widely used will be the `UnitaryMatrix` object which
represents a pure quantum operation in matrix form.

.. rubric:: Abstract Base Classes

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/class_no_init.rst

    Unitary
    DifferentiableUnitary
    LocallyOptimizableUnitary
    StateVectorMap

.. rubric:: Core Classes

.. autosummary::
    :toctree: autogen
    :recursive:

    UnitaryMatrix
    UnitaryBuilder
    StateVector
    StateSystem
    PauliMatrices
    PermutationMatrix
    CouplingGraph

.. rubric:: Type Aliases

.. autosummary::
    :toctree: autogen
    :recursive:

    UnitaryLike
    StateLike
    StateSystemLike
    RealVector
    CouplingGraphLike
"""
from __future__ import annotations

from typing import Any

from bqskit.qis.pauli import PauliMatrices
from bqskit.qis.permutation import PermutationMatrix
from bqskit.qis.state import StateLike
from bqskit.qis.state import StateVector
from bqskit.qis.state import StateVectorMap
from bqskit.qis.state.system import StateSystem
from bqskit.qis.state.system import StateSystemLike
from bqskit.qis.unitary import DifferentiableUnitary
from bqskit.qis.unitary import LocallyOptimizableUnitary
from bqskit.qis.unitary import RealVector
from bqskit.qis.unitary import Unitary
from bqskit.qis.unitary import UnitaryBuilder
from bqskit.qis.unitary import UnitaryLike
from bqskit.qis.unitary import UnitaryMatrix


def __getattr__(name: str) -> Any:
    # Lazy imports # TODO: Fix circular imports in qis.graph properly
    if name == 'CouplingGraph':
        from bqskit.qis.graph import CouplingGraph
        return CouplingGraph

    if name == 'CouplingGraphLike':
        from bqskit.qis.graph import CouplingGraphLike
        return CouplingGraphLike

    raise AttributeError(f'module {__name__} has no attribute {name}')


__all__ = [
    'Unitary',
    'DifferentiableUnitary',
    'LocallyOptimizableUnitary',
    'UnitaryBuilder',
    'UnitaryMatrix',
    'StateVector',
    'UnitaryLike',
    'StateVectorMap',
    'PauliMatrices',
    'StateLike',
    'PermutationMatrix',
    'RealVector',
    'CouplingGraph',
    'CouplingGraphLike',
    'StateSystem',
    'StateSystemLike',
]



"""This module implements the StaticPlacementPass class."""
from __future__ import annotations

import logging
import time

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.qis.graph import CouplingGraph

_logger = logging.getLogger(__name__)


class StaticPlacementPass(BasePass):
    """Find a subgraph monomorphic to the coupling graph so that no SWAPs are
    needed."""

    def __init__(self, timeout_sec: float = 10) -> None:
        self.timeout_sec = timeout_sec

    def _find_monomorphic_subgraph(
        self,
        time_limit: float,
        physical_graph: CouplingGraph,
        num_logical_qudits: int,
        minimal_degrees: list[int],
        connected_indices: list[list[int]],
        current_placement: list[int] = [],
        current_index: int = 0,
    ) -> list[int]:
        """Recursively find a monomorphic subgraph."""
        if current_index == num_logical_qudits:
            return current_placement

        if time.time() > time_limit:
            return []

        # Find all possible placements for the current logical qudit
        candidate_indices = set()

        # Filter out occupied qudits and qudits with insufficient degrees
        physical_degrees = physical_graph.get_qudit_degrees()
        for x in range(physical_graph.num_qudits):
            if (
                physical_degrees[x] >= minimal_degrees[current_index]
                and x not in current_placement
            ):
                candidate_indices.add(x)

        # Filter out qudits that are not connected to previous logical qudits
        for i in connected_indices[current_index]:
            candidate_indices &= set(
                physical_graph.get_neighbors_of(current_placement[i]),
            )

        # Try all possible placements for the current logical qudit
        for x in candidate_indices:
            new_placement = current_placement + [x]
            result = self._find_monomorphic_subgraph(
                time_limit,
                physical_graph,
                num_logical_qudits,
                minimal_degrees,
                connected_indices,
                new_placement,
                current_index + 1,
            )
            if len(result) == num_logical_qudits:
                return result

        # If no valid placement is found, return an empty list
        return []

    def find_monomorphic_subgraph(
        self,
        physical_graph: CouplingGraph,
        logical_graph: CouplingGraph,
    ) -> list[int]:
        """Try all possible placements."""

        # To be optimized later
        logical_qubit_order = list(range(logical_graph.num_qudits))

        minimum_degrees = [
            logical_graph.get_qudit_degrees()[i] for i in logical_qubit_order
        ]
        connected_indices: list[list[int]] = [
            [] for _ in range(logical_graph.num_qudits)
        ]
        for i in range(logical_graph.num_qudits):
            for j in range(i):
                if logical_qubit_order[j] in logical_graph.get_neighbors_of(
                    logical_qubit_order[i],
                ):
                    connected_indices[i].append(j)

        # Find a monomorphic subgraph
        start_time = time.time()
        index_to_physical = self._find_monomorphic_subgraph(
            start_time + self.timeout_sec,
            physical_graph,
            logical_graph.num_qudits,
            minimum_degrees,
            connected_indices,
        )
        _logger.info(f'elapsed time: {time.time() - start_time}')
        if len(index_to_physical) == 0:
            return []

        # Convert the result to a placement
        placement = [-1] * logical_graph.num_qudits
        for i, x in enumerate(logical_qubit_order):
            placement[x] = index_to_physical[i]
        return placement

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        physical_graph = data.model.coupling_graph
        logical_graph = circuit.coupling_graph

        # Find an monomorphic subgraph
        placement = self.find_monomorphic_subgraph(
            physical_graph, logical_graph,
        )

        # Set the placement if it is valid
        if len(placement) == logical_graph.num_qudits and all(
            placement[e[1]] in physical_graph.get_neighbors_of(placement[e[0]])
            for e in logical_graph
        ):
            data.placement = placement
            _logger.info(f'Placed qudits on {data.placement}')
        else:
            _logger.info('No valid placement found')



"""This module implements the U3Gate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class U3Gate(QubitGate, DifferentiableUnitary, CachedClass, GeneralGate):
    """
    The U3 single qubit gate.

    It is given by the following parameterized unitary:

    .. math::

        \\begin{pmatrix}
        \\cos{\\frac{\\theta_0}{2}} &
        -\\exp({i\\theta_2})\\sin{\\frac{\\theta_0}{2}} \\\\
        \\exp({i\\theta_1})\\sin{\\frac{\\theta_0}{2}} &
        \\exp({i(\\theta_1 + \\theta_2)})\\cos{\\frac{\\theta_0}{2}} \\\\
        \\end{pmatrix}
    """

    _num_qudits = 1
    _num_params = 3
    _qasm_name = 'u3'

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        ct = np.cos(params[0] / 2)
        st = np.sin(params[0] / 2)
        cp = np.cos(params[1])
        sp = np.sin(params[1])
        cl = np.cos(params[2])
        sl = np.sin(params[2])
        el = cl + 1j * sl
        ep = cp + 1j * sp

        return UnitaryMatrix(
            [
                [ct, -el * st],
                [ep * st, ep * el * ct],
            ],
        )

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        ct = np.cos(params[0] / 2)
        st = np.sin(params[0] / 2)
        cp = np.cos(params[1])
        sp = np.sin(params[1])
        cl = np.cos(params[2])
        sl = np.sin(params[2])
        el = cl + 1j * sl
        ep = cp + 1j * sp
        del_ = -sl + 1j * cl
        dep_ = -sp + 1j * cp

        return np.array(
            [
                [  # wrt params[0]
                    [-0.5 * st, -0.5 * ct * el],
                    [0.5 * ct * ep, -0.5 * st * el * ep],
                ],

                [  # wrt params[1]
                    [0, 0],
                    [st * dep_, ct * el * dep_],
                ],

                [  # wrt params[2]
                    [0, -st * del_],
                    [0, ct * ep * del_],
                ],
            ], dtype=np.complex128,
        )

    def calc_params(self, utry: UnitaryMatrix) -> list[float]:
        """
        Calculate the three parameters to a U3Gate from a given unitary.

        Args:
            utry (UnitaryMatrix): The single-qubit unitary matrix to
                calculate a U3Gate's parameters for.

        Returns:
            list[float]: The three parameters to a U3Gate,
                such that, `U3Gate().get_unitary` will return `utry`.

        Raises:
            ValueError: If `utry` is not a single-qubit unitary.
        """
        if utry.radixes != (2,):
            raise ValueError('Expected single-qubit unitary.')

        mag = np.linalg.det(utry.numpy) ** (-1 / 2)
        special_utry = mag * utry
        a = np.angle(special_utry[1, 1])
        b = np.angle(special_utry[1, 0])
        c = np.abs(special_utry[1, 0])
        d = np.abs(special_utry[0, 0])
        theta = 2 * float(np.arctan2(c, d))
        phi = (a + b)
        lamb = (a - b)
        return [theta, phi, lamb]

    def get_inverse_params(self, params: RealVector = []) -> RealVector:
        """Return the parameters that invert the gate, see :class:`Gate`."""
        self.check_parameters(params)
        return [-params[0], -params[2], -params[1]]

    def get_inverse(self) -> Gate:
        """Return the inverse of this gate."""
        return U3Gate()



"""This module implements the ISwapGate."""
from __future__ import annotations

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class ISwapGate(ConstantGate, QubitGate):
    """
    The two qubit swap and phase iSWAP gate.

    The ISwap gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 0 & i & 0 \\\\
        0 & i & 0 & 0 \\\\
        0 & 0 & 0 & 1 \\\\
        \\end{pmatrix}
    """

    _num_qudits = 2
    _qasm_name = 'iswap'
    _utry = UnitaryMatrix(
        [
            [1, 0, 0, 0],
            [0, 0, 1j, 0],
            [0, 1j, 0, 0],
            [0, 0, 0, 1],
        ],
    )

    def get_qasm_gate_def(self) -> str:
        """Returns a qasm gate definition block for this gate."""
        return (
            'gate iswap a,b\n'
            '{\n'
            '\ts a;\n'
            '\ts b;\n'
            '\th a;\n'
            '\tcx a, b;\n'
            '\tcx b, a;\n'
            '\th b;\n'
            '}\n'
        )



"""This module implements the QFactor class."""
from __future__ import annotations

import copy
import logging
from typing import Any
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from bqskitrs import QFactorInstantiatorNative

from bqskit.ir.opt.instantiater import Instantiater
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class QFactor(QFactorInstantiatorNative, Instantiater):
    """The QFactor circuit instantiater."""

    def __new__(cls, **kwargs: dict[str, Any]) -> Any:
        if 'cost_fn_gen' in kwargs:
            del kwargs['cost_fn_gen']

        if 'ftol' in kwargs:
            del kwargs['ftol']

        if 'gtol' in kwargs:
            del kwargs['gtol']

        new_obj = super().__new__(cls, **kwargs)

        new_obj.__orig_kwargs = copy.deepcopy(kwargs)

        return new_obj

    def instantiate(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
        x0: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """Instantiate `circuit`, see Instantiater for more info."""
        return QFactorInstantiatorNative.instantiate(self, circuit, target, x0)

    @staticmethod
    def is_capable(circuit: Circuit) -> bool:
        """Return true if the circuit can be instantiated."""
        return all(
            isinstance(gate, LocallyOptimizableUnitary)
            for gate in circuit.gate_set
        )

    @staticmethod
    def get_violation_report(circuit: Circuit) -> str:
        """
        Return a message explaining why `circuit` cannot be instantiated.

        Args:
            circuit (Circuit): Generate a report for this circuit.

        Raises:
            ValueError: If `circuit` can be instantiated with this
                instantiater.
        """

        invalid_gates = {
            gate
            for gate in circuit.gate_set
            if not isinstance(gate, LocallyOptimizableUnitary)
        }

        if len(invalid_gates) == 0:
            raise ValueError('Circuit can be instantiated.')

        return (
            'Cannot instantiate circuit with qfactor'
            ' because the following gates are not locally optimizable: %s.'
            % ', '.join(str(g) for g in invalid_gates)
        )

    @staticmethod
    def get_method_name() -> str:
        """Return the name of this method."""
        return 'qfactor'

    def __getnewargs_ex__(self) -> tuple[Any, Any]:
        return (tuple(), self.__orig_kwargs)



"""This module implements the CSUMDGate."""
from __future__ import annotations

import numpy as np

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class CSUMGate(ConstantGate, QuditGate):
    """
    The two-qudit Conditional-SUM gate.

    The CSUM gate is given by the following formula:

    .. math::
        CSUM |i,j\\rangle = |i, i + j \\mathrm{mod} d\\rangle

    where d is the number of levels (2 levels is a qubit,
    3 levels is a qutrit, etc.)


    For qutrits the CSUMGate is represented by the following matrix:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\\\
        0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 \\\\
        0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\\\
        \\end{pmatrix}

    References:
        - https://www.frontiersin.org/articles/10.3389/fphy.2020.589504/full
        - https://pubs.aip.org/aip/jmp/article-abstract/56/3/032202/763827
    """

    _num_qudits = 2
    _num_params = 0

    def __init__(self, radix: int = 3) -> None:
        """
        Construct a CSUMGate.

        Args:
            radix (int): The number of qudit levels (>=2). Defaults to
                qutrits.

        Raises:
            ValueError: if radix < 2
        """
        if not is_integer(radix):
            raise TypeError(f'Expected integer for radix, got {type(radix)}.')

        if radix < 2:
            raise ValueError(f'Radix must be greater than 1, got {radix}.')

        self._radix = radix

        matrix = np.zeros([self.radix**2, self.radix**2])
        for i in range(self.radix):
            for j in range(self.radix):
                row = self.radix * i + ((i + j) % self.radix)
                col = self.radix * i + j
                matrix[row, col] = 1.0
        self._utry = UnitaryMatrix(matrix, self.radixes)



"""This module implements the GateCountPredicate class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.ir.gate import Gate
from bqskit.passes.control.predicate import PassPredicate
from bqskit.utils.typing import is_sequence

if TYPE_CHECKING:
    from typing import Sequence

    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class GateCountPredicate(PassPredicate):
    """
    The GateCountPredicate class.

    The GateCountPredicate returns true if the number of one of more types of
    gate has changed in the circuit.
    """

    key = 'GateCountPredicate_circuit_count'

    def __init__(self, gate: Gate | Sequence[Gate] | str) -> None:
        """
        Construct a GateCountPredicate.

        Args:
            gate (Gate | Sequence[Gate] | str): The gate(s) to count.
             Either a single gate, a sequence of gates, or a string
                representing the gate type. The allowed strings are:
                    - 'sq' for single qudit gates
                    - 'tq' for two qudit gates
                    - 'multi' for multi-qudit gates
                    - 'many' for gates with more than 2 qudits
        """

        if isinstance(gate, Gate):
            gate = [gate]

        if not isinstance(gate, str):
            if not is_sequence(gate):
                raise TypeError(f'Expected gate, got {type(gate)}')

            if not all(isinstance(g, Gate) for g in gate):
                raise TypeError(f'Expected gate, got {type(gate)}')

        else:
            if gate not in ['sq', 'tq', 'multi', 'many']:
                raise ValueError(f'Unknown gate type {gate}.')

        self.gate = list(gate) if not isinstance(gate, str) else gate

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        if isinstance(self.gate, str):
            gates = list({
                'sq': circuit.gate_set.single_qudit_gates,
                'tq': circuit.gate_set.two_qudit_gates,
                'multi': circuit.gate_set.multi_qudit_gates,
                'many': circuit.gate_set.many_qudit_gates,
            }[self.gate])

        else:
            gates = self.gate  # type: ignore

        gate_count = sum(circuit.count(g) for g in gates)

        # If first call, record data and return true
        if self.key not in data:
            _logger.debug(f'Could not find {self.key} key.')
            data[self.key] = gate_count
            return True

        # Otherwise, check for a change and return accordingly
        if data[self.key] == gate_count:
            _logger.debug('Counts match; no change detected.')
            return False

        data[self.key] = gate_count
        _logger.debug('Counts do not match; change detected.')
        return True



"""This module implements the GeneralGate base class."""
from __future__ import annotations

import abc
from typing import Sequence

import numpy as np
import numpy.typing as npt
import scipy as sp

from bqskit.ir.gate import Gate
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class GeneralGate(Gate, LocallyOptimizableUnitary):
    """An abstract base class for gates that parameterize any unitary."""

    @abc.abstractmethod
    def calc_params(self, utry: UnitaryMatrix) -> list[float]:
        """Return the parameters for this gate to implement `utry`."""

    def identity_as_params(self, radixes: Sequence[int]) -> list[float]:
        """Return the parameters for the gate that implements the identity."""
        identity = UnitaryMatrix.identity(int(np.prod(radixes)), radixes)
        return self.calc_params(identity)

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        self.check_env_matrix(env_matrix)
        U, _, Vh = sp.linalg.svd(env_matrix)
        new_U = Vh.conj().T @ U.conj().T
        return self.calc_params(UnitaryMatrix(new_U, self.radixes))



"""This module implements the LogPass and LogErrorPass classes."""
from __future__ import annotations

import logging

import numpy as np

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class LogPass(BasePass):
    """A pass that logs a message to the BQSKit log."""

    def __init__(self, msg: str, level: int = logging.INFO) -> None:
        """
        Construct a LogPass.

        Args:
            msg (str): The message to log.

            level (int): The logging verbosity level.
        """
        self.msg = msg
        self.level = level

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.log(self.level, self.msg)


class LogErrorPass(BasePass):
    """A pass that logs the current error to the BQSKit log."""

    def __init__(self, error_threshold: float | None = None) -> None:
        """
        Construct a LogErrorPass.

        Args:
            error_threshold (float | None): Logs a warning if the error
                is above this threshold. If None, never log a warning.
                (Default: None)
        """
        self.threshold = error_threshold

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        if 'error' in data:
            error = data['error']
            nonsq_error = 1 - np.sqrt(max(1 - (error * error), 0))
            if self.threshold is not None and nonsq_error > self.threshold:
                _logger.warn(
                    'Upper bound on error is greater than set threshold:'
                    f' {nonsq_error} > {self.threshold}.',
                )
            else:
                _logger.info(f'Upper bound on error is {nonsq_error}.')



"""This module implements the SetTargetPass class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.compiler.basepass import BasePass
from bqskit.qis.state import StateSystem
from bqskit.qis.state import StateVector
from bqskit.qis.unitary import UnitaryMatrix

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


class SetTargetPass(BasePass):
    """Sets a synthesis target for future passes."""

    def __init__(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> None:
        """
        Construct a SetTargetPass.

        Args:
            target (UnitaryMatrix | StateVector | StateSystem): The target
                to synthesize in future passes.
        """
        if not isinstance(target, (StateVector, UnitaryMatrix, StateSystem)):
            bad_type = type(target)
            m = f'Expected valid unitary or state for target, got {bad_type}.'
            raise TypeError(m)

        self.target = target

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        data.target = self.target



"""
=============================================================
BQSKit Extensions (:mod:`bqskit.ext`)
=============================================================

.. currentmodule:: bqskit.ext

This subpackage provides integrations with other popular frameworks
and pre-built MachineModels for many QPUs. For some, you will need to
manually install the dependencies. For example, to use the Pytket translators
you will need to install the `pytket` package.

.. rubric:: Pre-Built Machine Models

.. autosummary::
    :toctree: autogen
    :recursive:

    Aspen11Model
    AspenM2Model
    ANKAA2Model
    ANKAA9Q3Model
    H1_1Model
    H1_2Model
    H2_1Model
    Sycamore23Model
    SycamoreModel
    model_from_backend

.. rubric:: Translators

.. autosummary::
    :toctree: autogen
    :recursive:

    bqskit_to_cirq
    cirq_to_bqskit
    bqskit_to_pytket
    pytket_to_bqskit
    bqskit_to_qiskit
    qiskit_to_bqskit
    bqskit_to_qutip
    qutip_to_bqskit

.. rubric:: SupermarQ Metrics

.. autosummary::
    :toctree: autogen
    :recursive:

    supermarq_program_communication
    supermarq_critical_depth
    supermarq_entanglement_ratio
    supermarq_parallelism
    supermarq_liveness
"""
from __future__ import annotations

from bqskit.ext.cirq.models import Sycamore23Model
from bqskit.ext.cirq.models import SycamoreModel
from bqskit.ext.cirq.translate import bqskit_to_cirq
from bqskit.ext.cirq.translate import cirq_to_bqskit
from bqskit.ext.pytket.translate import bqskit_to_pytket
from bqskit.ext.pytket.translate import pytket_to_bqskit
from bqskit.ext.qiskit.models import model_from_backend
from bqskit.ext.qiskit.translate import bqskit_to_qiskit
from bqskit.ext.qiskit.translate import qiskit_to_bqskit
from bqskit.ext.quantinuum import H1_1Model
from bqskit.ext.quantinuum import H1_2Model
from bqskit.ext.quantinuum import H2_1Model
from bqskit.ext.qutip.translate import bqskit_to_qutip
from bqskit.ext.qutip.translate import qutip_to_bqskit
from bqskit.ext.rigetti import ANKAA2Model
from bqskit.ext.rigetti import ANKAA9Q3Model
from bqskit.ext.rigetti import Aspen11Model
from bqskit.ext.rigetti import AspenM2Model
from bqskit.ext.supermarq import supermarq_critical_depth
from bqskit.ext.supermarq import supermarq_entanglement_ratio
from bqskit.ext.supermarq import supermarq_liveness
from bqskit.ext.supermarq import supermarq_parallelism
from bqskit.ext.supermarq import supermarq_program_communication
# TODO: Deprecate imports from __init__, use lazy import to deprecate


__all__ = [
    'bqskit_to_cirq',
    'cirq_to_bqskit',
    'bqskit_to_pytket',
    'pytket_to_bqskit',
    'model_from_backend',
    'bqskit_to_qiskit',
    'qiskit_to_bqskit',
    'bqskit_to_qutip',
    'qutip_to_bqskit',
    'supermarq_program_communication',
    'supermarq_critical_depth',
    'supermarq_entanglement_ratio',
    'supermarq_parallelism',
    'supermarq_liveness',
    'Aspen11Model',
    'AspenM2Model',
    'H1_1Model',
    'H1_2Model',
    'H2_1Model',
    'ANKAA2Model',
    'ANKAA9Q3Model',
    'Sycamore23Model',
    'SycamoreModel',
]



"""This module implements the LayerGenerator base class."""
from __future__ import annotations

import abc

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class LayerGenerator(abc.ABC):
    """
    The LayerGenerator base class.

    Search based synthesis uses the layer generator to generate the root node
    and the successors of a node.
    """

    @abc.abstractmethod
    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """Generate the initial layer for search."""

    @abc.abstractmethod
    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """Generate the successors of a circuit node."""



"""This module implements the Frontier class."""
from __future__ import annotations

import heapq
import itertools
from typing import Any
from typing import NamedTuple

from bqskit.ir.circuit import Circuit
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class FrontierElement(NamedTuple):
    """The Frontier contains FrontierElements."""
    cost: float
    element_id: int
    circuit: Circuit
    extra_data: Any


class Frontier:
    """The Frontier class."""

    def __init__(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        heuristic_function: HeuristicFunction,
    ) -> None:
        """
        Construct an empty frontier.

        Args:
            target (UnitaryMatrix | StateVector | StateSystem): The target to
                pass to the heuristic_function.

            heuristic_function (HeuristicFunction): The heuristic used
                to sort the Frontier.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        if not isinstance(heuristic_function, HeuristicFunction):
            raise TypeError(
                'Expected HeursiticFunction, got %s.'
                % type(heuristic_function),
            )

        self.target = target
        self.heuristic_function = heuristic_function
        self._frontier: list[FrontierElement] = []
        self._counter = itertools.count()

    def add(self, circuit: Circuit, extra_data: Any = None) -> None:
        """Add `circuit` into the frontier."""
        heuristic_value = self.heuristic_function(circuit, self.target)
        count = next(self._counter)
        elem = FrontierElement(heuristic_value, count, circuit, extra_data)
        heapq.heappush(self._frontier, elem)

    def pop(self) -> tuple[Circuit, Any]:
        """Pop the top circuit."""
        elem = heapq.heappop(self._frontier)
        return elem.circuit, elem.extra_data

    def empty(self) -> bool:
        """Return true if the frontier is empty."""
        return len(self._frontier) == 0

    def clear(self) -> None:
        """Remove all elements from the frontier."""
        self._frontier.clear()



"""This module implements the GroupSingleQuditGatePass."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import MeasurementPlaceholder
from bqskit.ir.gates import Reset
from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.region import CircuitRegion


class GroupSingleQuditGatePass(BasePass):
    """
    The GroupSingleQuditGatePass Pass.

    This pass groups together consecutive single-qudit gates.
    """

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Go through each qudit individually
        for q in range(circuit.num_qudits):

            single_qubit_regions = []
            region_start = None

            for c in range(circuit.num_cycles):
                if circuit.is_point_idle((c, q)):
                    continue

                op = circuit[c, q]
                if (
                    op.num_qudits == 1
                    and not isinstance(
                        op.gate, (
                            BarrierPlaceholder,
                            MeasurementPlaceholder,
                            Reset,
                        ),
                    )
                ):
                    if region_start is None:
                        region_start = c
                else:
                    if region_start is not None:
                        region = CircuitRegion({q: (region_start, c - 1)})
                        single_qubit_regions.append(region)
                        region_start = None

            if region_start is not None:
                region = CircuitRegion(
                    {q: (region_start, circuit.num_cycles - 1)},
                )
                single_qubit_regions.append(region)
                region_start = None

            for region in reversed(single_qubit_regions):
                circuit.fold(region)



"""This module implements the SynthesisPass abstract class."""
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class SynthesisPass(BasePass):
    """
    SynthesisPass abstract class.

    The SynthesisPass is a base class that exposes an abstract synthesize
    function. Inherit from this class and implement the synthesize function to
    create a synthesis tool.

    A SynthesisPass will synthesize a new circuit targeting the input circuit's
    unitary.
    """

    @abstractmethod
    async def synthesize(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Synthesis abstract method to synthesize a UnitaryMatrix into a Circuit.

        Args:
            utry (UnitaryMatrix): The unitary to synthesize.

            data (Dict[str, Any]): Associated data for the pass.
                Can be used to provide auxillary information from
                previous passes. This function should never error based
                on what is in this dictionary.

        Note:
            This function should be self-contained and have no side effects.
        """

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        circuit.become(await self.synthesize(data.target, data))



"""This module implements the NOOPPass class."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit


class NOOPPass(BasePass):
    """A pass that does not perform any operation."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        return



"""This module implements the QuditGate base class."""
from __future__ import annotations

from bqskit.ir.gate import Gate


class QuditGate(Gate):
    """A gate that acts on qudits, all with the same level."""

    _radix: int

    @property
    def radixes(self) -> tuple[int, ...]:
        """The number of orthogonal states for each qudit."""
        return tuple([self.radix] * self.num_qudits)

    @property
    def radix(self) -> int:
        """The number of levels in each qudit."""
        return getattr(self, '_radix')



"""This module implements the WalshDiagonalSynthesisPass."""
from __future__ import annotations

import logging

from numpy import where

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import CNOTGate
from bqskit.ir.gates import RZGate
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary import UnitaryMatrix
from bqskit.utils.math import pauliz_expansion
from bqskit.utils.math import unitary_log_no_i


_logger = logging.getLogger(__name__)


class WalshDiagonalSynthesisPass(SynthesisPass):
    """
    A pass that synthesizes diagonal unitaries into Walsh functions.

    Based on: https://arxiv.org/abs/1306.3991
    """

    def __init__(
        self,
        parameter_precision: float = 1e-8,
    ) -> None:
        """
        Constructor for WalshDiagonalSynthesisPass.

        Args:
            parameter_precision (float): Pauli strings with parameter values
                less than this are rounded to zero. (Default: 1e-8)

        TODO:
            - Cancel adjacent CNOTs
            - See how QFAST can be used to generalize to qudits
        """
        self.parameter_precision = parameter_precision

    def gray_code(self, number: int) -> int:
        """Convert a number to its Gray code representation."""
        gray = number ^ (number >> 1)
        return gray

    def pauli_to_subcircuit(
        self,
        string_id: int,
        angle: float,
        num_qubits: int,
    ) -> Circuit:
        string = bin(string_id)[2:].zfill(num_qubits)
        circuit = Circuit(num_qubits)
        locations = [i for i in range(num_qubits) if string[i] == '1']
        if len(locations) == 1:
            circuit.append_gate(RZGate(), locations[0], [angle])
        elif len(locations) > 1:
            pairs = [
                (locations[i], locations[i + 1])
                for i in range(len(locations) - 1)
            ]
            for pair in pairs:
                circuit.append_gate(CNOTGate(), pair)
            circuit.append_gate(RZGate(), locations[-1], [angle])
            for pair in reversed(pairs):
                circuit.append_gate(CNOTGate(), pair)
        return circuit

    async def synthesize(
        self,
        utry: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """Synthesize `utry`, see :class:`SynthesisPass` for more."""
        if not isinstance(utry, UnitaryMatrix):
            m = 'WalshDiagonalSynthesisPass can only synthesize diagonal, '
            m += f'`UnitaryMatrix`s, got {type(utry)}.'
            raise TypeError(m)

        if not utry.is_qubit_only():
            m = 'WalshDiagonalSynthesisPass can only synthesize diagonal '
            m += '`UnitaryMatrix`s with qubits, got higher radix than 2.'
            raise ValueError(m)

        num_qubits = utry.num_qudits
        circuit = Circuit(num_qubits)

        # Find parameters of each I/Z Pauli string
        H_matrix = unitary_log_no_i(utry.numpy)
        params = pauliz_expansion(H_matrix) * 2
        # Remove low weight terms - these are likely numerical errors
        params = where(abs(params) < self.parameter_precision, 0, params)

        # Order the Pauli strings by their Gray code representation
        pauli_params = sorted(
            [(i, -p) for i, p in enumerate(params)],
            key=lambda x: self.gray_code(x[0]),
        )
        subcircuits = [
            self.pauli_to_subcircuit(i, p, num_qubits) for i, p in pauli_params
        ]

        for subcircuit in subcircuits:
            circuit.append_circuit(subcircuit, [_ for _ in range(num_qubits)])

        return circuit



"""This package defines constant gates."""
from __future__ import annotations

from bqskit.ir.gates.constant.b import BGate
from bqskit.ir.gates.constant.ccx import CCXGate
from bqskit.ir.gates.constant.ccx import ToffoliGate
from bqskit.ir.gates.constant.ch import CHGate
from bqskit.ir.gates.constant.clock import ClockGate
from bqskit.ir.gates.constant.cpi import CPIGate
from bqskit.ir.gates.constant.cs import CSGate
from bqskit.ir.gates.constant.csum import CSUMGate
from bqskit.ir.gates.constant.ct import CTGate
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.gates.constant.cx import CXGate
from bqskit.ir.gates.constant.cy import CYGate
from bqskit.ir.gates.constant.cz import CZGate
from bqskit.ir.gates.constant.ecr import ECRGate
from bqskit.ir.gates.constant.h import HGate
from bqskit.ir.gates.constant.identity import IdentityGate
from bqskit.ir.gates.constant.iswap import ISwapGate
from bqskit.ir.gates.constant.itoffoli import IToffoliGate
from bqskit.ir.gates.constant.pd import PDGate
from bqskit.ir.gates.constant.permutation import PermutationGate
from bqskit.ir.gates.constant.rccx import MargolusGate
from bqskit.ir.gates.constant.rccx import RC3XGate
from bqskit.ir.gates.constant.rccx import RCCXGate
from bqskit.ir.gates.constant.s import SGate
from bqskit.ir.gates.constant.sdg import SdgGate
from bqskit.ir.gates.constant.shift import ShiftGate
from bqskit.ir.gates.constant.sqrtcnot import SqrtCNOTGate
from bqskit.ir.gates.constant.sqrtiswap import SqrtISwapGate
from bqskit.ir.gates.constant.sqrtt import SqrtTGate
from bqskit.ir.gates.constant.subswap import SubSwapGate
from bqskit.ir.gates.constant.swap import SwapGate
from bqskit.ir.gates.constant.sx import SqrtXGate
from bqskit.ir.gates.constant.sx import SXGate
from bqskit.ir.gates.constant.sycamore import SycamoreGate
from bqskit.ir.gates.constant.t import TGate
from bqskit.ir.gates.constant.tdg import TdgGate
from bqskit.ir.gates.constant.unitary import ConstantUnitaryGate
from bqskit.ir.gates.constant.x import XGate
from bqskit.ir.gates.constant.xx import XXGate
from bqskit.ir.gates.constant.y import YGate
from bqskit.ir.gates.constant.yy import YYGate
from bqskit.ir.gates.constant.z import ZGate
from bqskit.ir.gates.constant.zz import ZZGate

__all__ = [
    'BGate',
    'CCXGate',
    'ToffoliGate',
    'CHGate',
    'ClockGate',
    'CPIGate',
    'CSGate',
    'CSUMGate',
    'CTGate',
    'CNOTGate',
    'CXGate',
    'CYGate',
    'CZGate',
    'ECRGate',
    'HGate',
    'IdentityGate',
    'ISwapGate',
    'IToffoliGate',
    'PDGate',
    'PermutationGate',
    'MargolusGate',
    'RC3XGate',
    'RCCXGate',
    'SGate',
    'SdgGate',
    'ShiftGate',
    'SqrtCNOTGate',
    'SqrtISwapGate',
    'SqrtTGate',
    'SubSwapGate',
    'SwapGate',
    'SqrtXGate',
    'SXGate',
    'SycamoreGate',
    'TGate',
    'TdgGate',
    'ConstantUnitaryGate',
    'XGate',
    'XXGate',
    'YGate',
    'YYGate',
    'ZGate',
    'ZZGate',
]



"""
This module implements the Gate base class.

A gate is a potentially-parameterized, immutable, unitary operation that can be
applied to a circuit.
"""
from __future__ import annotations

from typing import Callable
from typing import ClassVar
from typing import TYPE_CHECKING

from bqskit.ir.location import CircuitLocation
from bqskit.qis.unitary.unitary import Unitary

if TYPE_CHECKING:
    from bqskit.qis.unitary.unitary import RealVector
    from bqskit.ir.gates.composed.frozenparam import FrozenParameterGate


class Gate(Unitary):
    """Gate Base Class."""

    _name: str
    _qasm_name: str

    @property
    def name(self) -> str:
        """The name of this gate."""
        return getattr(self, '_name', self.__class__.__name__)

    @property
    def qasm_name(self) -> str:
        """The qasm identifier for this gate."""
        if not self.is_qubit_only():
            raise AttributeError('QASM only supports qubit gates.')

        return getattr(self, '_qasm_name')

    def get_qasm_gate_def(self) -> str:
        """Returns a qasm gate definition block for this gate."""
        if not self.is_qubit_only():
            raise AttributeError('QASM only supports qubit gates.')

        return ''

    def get_qasm(self, params: RealVector, location: CircuitLocation) -> str:
        """Returns the qasm string for this gate."""
        return '{}({}) q[{}];\n'.format(
            self.qasm_name,
            ', '.join([str(p) for p in params]),
            '], q['.join([str(q) for q in location]),
        ).replace('()', '')

    def get_inverse_params(self, params: RealVector = []) -> RealVector:
        """
        Return the parameters that invert the gate.

        Args:
            params (RealVector): The parameters of the gate to invert.

        Note:
            - The default implementation returns the same paramters because
              the default implementation of `Gate.get_inverse` returns a
              :class:`DaggerGate` wrapper of the gate. The wrapper will
              correctly handle the inversion. When overriding `get_inverse`,
              on a parameterized gate this method should be overridden
              as well.
        """
        return params

    def get_inverse(self) -> Gate:
        """Return the gate's inverse as a gate."""
        if self.is_constant() and self.is_self_inverse():
            return self

        from bqskit.ir.gates.composed import DaggerGate
        return getattr(self, '_inverse', DaggerGate(self))
        # TODO: Fill out inverse definitions throughout the gate library

    with_frozen_params: ClassVar[
        Callable[[Gate, dict[int, float]], FrozenParameterGate]
    ]
    with_all_frozen_params: ClassVar[
        Callable[[Gate, list[float]], FrozenParameterGate]
    ]

    def __repr__(self) -> str:
        return self.name



"""This module implements the Reset class."""
from __future__ import annotations

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class Reset(ConstantGate):
    """Pseudogate to reset/initialize the qudit to |0>."""

    def __init__(self, radix: int = 2) -> None:
        """
        Construct a Reset.

        Args:
            radix (int): the dimension of the qudit. (Default: 2)
        """
        self._num_qudits = 1
        self._qasm_name = 'reset'
        self._radixes = tuple([radix])
        self._num_params = 0

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        raise RuntimeError('Cannot compute unitary for a reset.')



"""This module implements the ToU3Pass."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.ir.point import CircuitPoint

_logger = logging.getLogger(__name__)


class ToU3Pass(BasePass):
    """Converts single-qubit general unitary gates to U3 Gates."""

    def __init__(self, convert_all_single_qubit_gates: bool = False) -> None:
        """
        Construct a ToU3Pass.

        Args:
            convert_all_single_qubit_gates (bool): Indicates wheter to convert
            only the general gates, or every single qubit gate.
        """

        self.convert_all_single_qubit_gates = convert_all_single_qubit_gates

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.debug('Converting single-qubit general gates to U3Gates.')
        for cycle, op in circuit.operations_with_cycles():
            if (
                op.radixes == (2,) and (
                    isinstance(op.gate, GeneralGate)
                    or self.convert_all_single_qubit_gates
                )
            ):
                params = U3Gate().calc_params(op.get_unitary())
                point = CircuitPoint(cycle, op.location[0])
                circuit.replace_gate(point, U3Gate(), op.location, params)



"""This package contains HeuristicFunction definitions."""
from __future__ import annotations

from bqskit.passes.search.heuristics.astar import AStarHeuristic
from bqskit.passes.search.heuristics.dijkstra import DijkstraHeuristic
from bqskit.passes.search.heuristics.greedy import GreedyHeuristic

__all__ = ['AStarHeuristic', 'GreedyHeuristic', 'DijkstraHeuristic']



"""This module implements the ClockGate."""
from __future__ import annotations

import numpy as np

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class ClockGate(ConstantGate, QuditGate):
    """
    The one-qudit clock (Z) gate. This is a Weyl-Heisenberg gate.

    The clock gate is given by the following formula:

    .. math::
        Z = \\sum_a \\exp(\\frac{2\\pi ai}{d}) |a\\rangle\\langle a|

    where d is the number of levels (2 levels is a qubit,
    3 levels is a qutrit, etc.)

    References:
        - https://arxiv.org/pdf/2302.07966.pdf
    """

    _num_qudits = 1

    def __init__(self, radix: int = 3) -> None:
        """
        Construct a ClockGate.

        Args:
            radix (int): The number of qudit levels (>=2). Defaults to
                qutrits.

        Raises:
            ValueError: if radix < 2
        """

        if not is_integer(radix):
            raise TypeError(f'Expected integer for radix, got {type(radix)}.')

        if radix < 2:
            raise ValueError(f'Radix must be greater than 1, got {radix}.')

        self._radix = radix

        # Calculate unitary
        diags = [np.exp(2j * np.pi * i / radix) for i in range(radix)]
        self._utry = UnitaryMatrix(np.diag(diags), self.radixes)



"""This module implements the QSearchSynthesisPass."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.passes.search.frontier import Frontier
from bqskit.passes.search.generator import LayerGenerator
from bqskit.passes.search.generators.seed import SeedLayerGenerator
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.passes.search.heuristics import AStarHeuristic
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary import UnitaryMatrix
from bqskit.runtime import get_runtime
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number


_logger = logging.getLogger(__name__)


class QSearchSynthesisPass(SynthesisPass):
    """
    A pass implementing the QSearch A* synthesis algorithm.

    References:
        Davis, Marc G., et al. “Towards Optimal Topology Aware Quantum
        Circuit Synthesis.” 2020 IEEE International Conference on Quantum
        Computing and Engineering (QCE). IEEE, 2020.
    """

    def __init__(
        self,
        heuristic_function: HeuristicFunction = AStarHeuristic(),
        layer_generator: LayerGenerator | None = None,
        success_threshold: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        max_layer: int | None = None,
        no_progress_layers_allowed: int = 10,
        store_partial_solutions: bool = False,
        partials_per_depth: int = 25,
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        Construct a search-based synthesis pass.

        Args:
            heuristic_function (HeuristicFunction): The heuristic to guide
                search.

            layer_generator (LayerGenerator | None): The successor function
                to guide node expansion. If left as none, then a default
                will be selected before synthesis based on the target
                model's gate set. (Default: None)

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the cost function. (Default: 1e-8)

            cost (CostFunction | None): The cost function that determines
                distance during synthesis. The goal of this synthesis pass
                is to implement circuits for the given unitaries that have
                a cost less than the `success_threshold`.
                (Default: HSDistance())

            max_layer (int): The maximum number of layers to append without
                success before termination. If left as None it will default
                to unlimited. (Default: None)

            no_progress_layers_allowed (int): The maximum number of layers
                allowed without improvement before a warning is issued to
                the user. (Default: 10)

            store_partial_solutions (bool): Whether to store partial solutions
                at different depths inside of the data dict. (Default: False)

            partials_per_depth (int): The maximum number of partials
                to store per search depth. No effect if
                `store_partial_solutions` is False. (Default: 25)

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

        Raises:
            ValueError: If `max_depth` is nonpositive.
        """
        if not isinstance(heuristic_function, HeuristicFunction):
            raise TypeError(
                f'Expected HeursiticFunction, got {type(heuristic_function)}.',
            )

        if layer_generator is not None:
            if not isinstance(layer_generator, LayerGenerator):
                raise TypeError(
                    f'Expected LayerGenerator, got {type(layer_generator)}.',
                )

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                f', got {type(success_threshold)}',
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator'
                f', got {type(cost)}',
            )

        if max_layer is not None and not is_integer(max_layer):
            raise TypeError(
                f'Expected max_layer to be an integer, got {type(max_layer)}.',
            )

        if max_layer is not None and max_layer <= 0:
            raise ValueError(
                f'Expected max_layer to be positive, got {int(max_layer)}.',
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options'
                f', got {type(instantiate_options)}',
            )

        self.heuristic_function = heuristic_function
        self.layer_gen = layer_generator
        self.success_threshold = success_threshold
        self.cost = cost
        self.max_layer = max_layer
        self.no_progress_layers_allowed = no_progress_layers_allowed
        self.instantiate_options: dict[str, Any] = {
            'cost_fn_gen': self.cost,
        }
        self.instantiate_options.update(instantiate_options)
        self.store_partial_solutions = store_partial_solutions
        self.partials_per_depth = partials_per_depth

    async def synthesize(
        self,
        utry: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """Synthesize `utry`, see :class:`SynthesisPass` for more."""
        # Initialize run-dependent options
        instantiate_options = self.instantiate_options.copy()

        # Seed the PRNG
        if 'seed' not in instantiate_options:
            instantiate_options['seed'] = data.seed

        # Get layer generator for search
        layer_gen = self._get_layer_gen(data)

        # Begin the search with an initial layer
        frontier = Frontier(utry, self.heuristic_function)
        initial_layer = layer_gen.gen_initial_layer(utry, data)
        initial_layer.instantiate(utry, **instantiate_options)
        frontier.add(initial_layer, 0)

        # Track best circuit, initially the initial layer
        best_dist = self.cost.calc_cost(initial_layer, utry)
        best_circ = initial_layer
        best_layer = 0

        # Track partial solutions
        psols: dict[int, list[tuple[Circuit, float]]] = {}

        _logger.debug(f'Search started, initial layer has cost: {best_dist}.')

        # Evalute initial layer
        if best_dist < self.success_threshold:
            _logger.debug('Successful synthesis with 0 layers.')
            return initial_layer

        # Record layers that have been warned about
        # to avoid duplicate warnings
        warned_layers: list[int] = []

        # Main loop
        while not frontier.empty():
            top_circuit, layer = frontier.pop()

            # Generate successors
            successors = layer_gen.gen_successors(top_circuit, data)

            if len(successors) == 0:
                continue

            # Instantiate successors
            circuits = await get_runtime().map(
                Circuit.instantiate,
                successors,
                target=utry,
                **instantiate_options,
            )

            # Evaluate successors
            for circuit in circuits:
                dist = self.cost.calc_cost(circuit, utry)

                if dist < self.success_threshold:
                    _logger.debug(
                        f'Successful synthesis with {layer + 1} layers.',
                    )
                    if self.store_partial_solutions:
                        data['psols'] = psols
                    return circuit

                if dist < best_dist:
                    plural = '' if layer == 0 else 's'
                    _logger.debug(
                        f'New best circuit found with {layer + 1} '
                        f'layer{plural} and cost: {dist:.12e}.',
                    )
                    best_dist = dist
                    best_circ = circuit
                    best_layer = layer + 1

                if self.store_partial_solutions:
                    if layer not in psols:
                        psols[layer] = []

                    psols[layer].append((circuit.copy(), dist))

                    if len(psols[layer]) > self.partials_per_depth:
                        psols[layer].sort(key=lambda x: x[1])
                        del psols[layer][-1]

                if self.max_layer is None or layer + 1 < self.max_layer:
                    frontier.add(circuit, layer + 1)

            layer_diff = abs(best_layer - layer)
            if (
                layer_diff % self.no_progress_layers_allowed == 0
                and layer_diff > 0
                and layer not in warned_layers
            ):
                _logger.warning(
                    'No improvement after '
                    f'{self.no_progress_layers_allowed} layers.',
                )
                warned_layers.append(layer)

        _logger.warning('Frontier emptied.')
        _logger.warning(
            'Returning best known circuit with %d layer%s and cost: %e.'
            % (best_layer, '' if best_layer == 1 else 's', best_dist),
        )
        if self.store_partial_solutions:
            data['psols'] = psols

        return best_circ

    def _get_layer_gen(self, data: PassData) -> LayerGenerator:
        """
        Set the layer generator.

        If a layer generator has been passed into the constructor, then that
        layer generator will be used. Otherwise, a default layer generator will
        be selected by the gateset.

        If seeds are passed into the data dict, then a SeedLayerGenerator will
        wrap the previously selected layer generator.
        """
        # TODO: Deduplicate this code with leap synthesis
        layer_gen = self.layer_gen or data.gate_set.build_mq_layer_generator()

        # Priority given to seeded synthesis
        if 'seed_circuits' in data:
            return SeedLayerGenerator(data['seed_circuits'], layer_gen)

        return layer_gen



"""This module implements the HeuristicFunction base class."""
from __future__ import annotations

import abc

from bqskit.ir.circuit import Circuit
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class HeuristicFunction(abc.ABC):
    """
    The HeuristicFunction base class.

    A HeuristicFunction is a map from a circuit to a real value.
    """

    @abc.abstractmethod
    def get_value(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> float:
        """Return the heuristic's value given `circuit` and `target`."""

    def __call__(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> float:
        """Return the heuristic's value given `circuit` and `target`."""

        if not isinstance(circuit, Circuit):
            raise TypeError(
                'Expected circuit, got %s.' % type(circuit),
            )

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        return self.get_value(circuit, target)



"""This module implements the Quantum Shannon Decomposition."""
from __future__ import annotations

import logging
from typing import Any

import numpy as np
from scipy.linalg import cossin
from scipy.linalg import schur

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.compiler.workflow import Workflow
from bqskit.ir.circuit import Circuit
from bqskit.ir.circuit import CircuitLocation
from bqskit.ir.circuit import CircuitPoint
from bqskit.ir.gates import CircuitGate
from bqskit.ir.gates.constant import CNOTGate
from bqskit.ir.gates.parameterized import RYGate
from bqskit.ir.gates.parameterized import RZGate
from bqskit.ir.gates.parameterized import VariableUnitaryGate
from bqskit.ir.gates.parameterized.mpry import MPRYGate
from bqskit.ir.gates.parameterized.mprz import MPRZGate
from bqskit.ir.operation import Operation
from bqskit.passes.processing import ScanningGateRemovalPass
from bqskit.passes.processing import TreeScanningGateRemovalPass
from bqskit.qis.permutation import PermutationMatrix
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.runtime import get_runtime

_logger = logging.getLogger(__name__)


class FullQSDPass(BasePass):
    """
    A pass performing one round of decomposition from the QSD algorithm.

    Important: This pass runs on VariableUnitaryGates only. Make sure to convert
    any gates you want to decompose to VariableUnitaryGates before running this
    pass.

    Additionally, ScanningGateRemovalPass will operate on the context of the
    entire circuit. If your circuit is large, it is best to set `perform_scan`
    to False.

    References:
        C.C. Paige, M. Wei,
        History and generality of the CS decomposition,
        Linear Algebra and its Applications,
        Volumes 208–209,
        1994,
        Pages 303-326,
        ISSN 0024-3795,
        https://doi.org/10.1016/0024-3795(94)90446-4.
    """

    def __init__(
        self,
        min_qudit_size: int = 2,
        perform_scan: bool = False,
        start_from_left: bool = True,
        tree_depth: int = 0,
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        Construct a single level of the QSDPass.

        Args:
            min_qudit_size (int): Performs QSD until the circuit only has
                VariableUnitaryGates with a number of qudits less than or equal
                to this value. (Default: 2)
            perform_scan (bool): Whether or not to perform the scanning
                gate removal pass. (Default: False)
            start_from_left (bool): Determines where the scan starts
                attempting to remove gates from. If True, scan goes left
                to right, otherwise right to left. (Default: True)
            tree_depth (int): The depth of the tree to use in the
                TreeScanningGateRemovalPass. If set to 0, we will instead
                use the ScanningGateRemovalPass. (Default: 0)
            instantiate_options (dict): The options to pass to the
                scanning gate removal pass. (Default: {})
        """
        self.start_from_left = start_from_left
        self.min_qudit_size = min_qudit_size
        instantiation_options = {'method': 'qfactor'}
        instantiation_options.update(instantiate_options)
        self.scan = ScanningGateRemovalPass(
            start_from_left=start_from_left,
            instantiate_options=instantiation_options,
        )
        if tree_depth > 0:
            self.scan = TreeScanningGateRemovalPass(
                start_from_left=start_from_left,
                instantiate_options=instantiation_options,
                tree_depth=tree_depth,
            )
        # Instantiate the helper QSD pass
        self.qsd = QSDPass(min_qudit_size=min_qudit_size)
        # Instantiate the helper Multiplex Gate Decomposition pass
        self.mgd = MGDPass(decompose_twice=False)
        self.perform_scan = perform_scan

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Run a round of QSD, Multiplex Gate Decomposition, and Scanning Gate
        Removal (optionally) until you reach the desired qudit size gates."""
        passes: list[BasePass] = []
        start_num = max(x.num_qudits for x in circuit.operations())
        for _ in range(self.min_qudit_size, start_num):
            passes.append(self.qsd)
            if self.perform_scan:
                passes.append(self.scan)
            passes.append(self.mgd)
        await Workflow(passes).run(circuit, data)


class MGDPass(BasePass):
    """
    A pass performing one round of decomposition of the MPRY and MPRZ gates in a
    circuit.

    References:
        C.C. Paige, M. Wei,
        History and generality of the CS decomposition,
        Linear Algebra and its Applications,
        Volumes 208–209,
        1994,
        Pages 303-326,
        ISSN 0024-3795,
        https://arxiv.org/pdf/quant-ph/0406176.pdf
    """

    def __init__(self, decompose_twice: bool = True) -> None:
        """
        The MGDPass decomposes all MPRY and MPRZ gates in a circuit.

        Args:
            decompose_twice (bool): Whether to decompose the MPRZ gate twice.
            This will save 2 CNOT gates in the decomposition. If false,
            the pass will only decompose one level. (Default: True)
        """
        self.decompose_twice = decompose_twice

    @staticmethod
    def decompose_mpx_one_level(
        decompose_ry: bool,
        params: RealVector,
        num_qudits: int,
        reverse: bool = False,
        drop_last_cnot: bool = False,
    ) -> Circuit:
        """
        Decompose Multiplexed Gate one level.

        Args:
            params (RealVector): The parameters for the original MPRZ gate
            num_qudits (int): The number of qudits in the MPRZ gate
            reverse (bool): Whether to reverse the order of the gates (you can
            decompose the gates in either order to get the same result)
            drop_last_cnot (bool): Whether to drop the last CNOT gate. This
            should be set if you are doing a 2 level decomposition to save 2
            CNOT gates.

        Returns:
            Circuit: The circuit that decomposes the MPRZ gate
        """

        new_gate: MPRZGate | MPRYGate | RZGate | RYGate = RZGate()
        if decompose_ry:
            new_gate = RYGate()

        if (num_qudits >= 3):
            if decompose_ry:
                new_gate = MPRYGate(num_qudits - 1, num_qudits - 2)
            else:
                # Remove 1 qubit, last qubit is controlled
                new_gate = MPRZGate(num_qudits - 1, num_qudits - 2)

        left_params, right_params = MPRYGate.get_decomposition(params)
        circ = Circuit(num_qudits)
        new_gate_location = list(range(1, num_qudits))
        cx_location = (0, num_qudits - 1)

        ops = [
            Operation(new_gate, new_gate_location, left_params),
            Operation(CNOTGate(), cx_location),
            Operation(new_gate, new_gate_location, right_params),
            Operation(CNOTGate(), cx_location),
        ]

        if drop_last_cnot:
            ops.pop()

        if reverse:
            ops.reverse()

        for op in ops:
            circ.append(op)

        return circ

    @staticmethod
    def decompose_mpx_two_levels(
        decompose_ry: bool,
        params: RealVector,
        num_qudits: int,
        reverse: bool = False,
        drop_last_cnot: bool = False,
    ) -> Circuit:
        """
        We decompose a multiplexed RZ gate 2 levels deep. This allows you to
        remove 2 CNOTs as per Figure 2 in
        https://arxiv.org/pdf/quant-ph/0406176.pdf.

        Furthermore, in the context of the Block ZXZ decomposition, you can
        set `drop_last_cnot` to True. This CNOT gets merged into a central gate,
        which saves another 2 CNOTs. This is shown in section 5.2 of
        https://arxiv.org/pdf/2403.13692v1.pdf.

        Args:
            decompose_ry (bool): Whether to decompose the MPRY gate
            params (RealVector): The parameters for the original MPR gate
            num_qudits (int): The number of qudits in the MPR gate
            reverse (bool): Whether to reverse the order of the gates (you can
            decompose the gates in either order to get the same result)
            drop_last_cnot (bool): Whether to drop the last CNOT gate (only
            should be set to True if you are doing section 5.2 optimization)

        Returns:
            Circuit: The circuit that decomposes the MPR gate
        """

        if num_qudits <= 2:
            # If you have less than 3 qubits, just decompose one level
            return MGDPass.decompose_mpx_one_level(
                decompose_ry,
                params,
                num_qudits,
                reverse,
            )

        # Get params for first decomposition of the MPRZ gate
        left_params, right_params = MPRYGate.get_decomposition(params)

        # Decompose the MPRZ gate into 2 MPRZ gates, dropping the last CNOT
        # Also Reverse the circuit for the right side in order to do the
        # optimization in section 5.2
        circ_left = MGDPass.decompose_mpx_one_level(
            decompose_ry,
            left_params,
            num_qudits - 1,
            reverse,
            drop_last_cnot=True,
        )

        circ_right = MGDPass.decompose_mpx_one_level(
            decompose_ry,
            right_params,
            num_qudits - 1,
            not reverse,
            drop_last_cnot=True,
        )

        # Now, construct the circuit.
        # This will generate the original MPRZ gate with the target qubit
        # set as qubit num_qudits - 1
        circ = Circuit(num_qudits)
        cx_location_big = (0, num_qudits - 1)

        ops: list[Circuit | Operation] = [
            circ_left,
            Operation(CNOTGate(), cx_location_big),
            circ_right,
            Operation(CNOTGate(), cx_location_big),
        ]

        # We can remove the last CNOT gate as per section 5.2
        if drop_last_cnot:
            ops.pop()

        if reverse:
            ops.reverse()

        for op in ops:
            if isinstance(op, Operation):
                circ.append(op)
            else:
                circ.append_circuit(op, list(range(1, num_qudits)))

        return circ

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Decompose all MPRY and MPRZ gates in the circuit one level."""
        ops: list[Operation] = []
        pts: list[CircuitPoint] = []
        locations: list[CircuitLocation] = []
        all_ops = list(circuit.operations_with_cycles(reverse=True))

        # Gather all of the multiplexed operations
        for cyc, op in all_ops:
            if isinstance(op.gate, MPRYGate) or isinstance(op.gate, MPRZGate):
                ops.append(op)
                pts.append(CircuitPoint((cyc, op.location[0])))
                # Adjust location based on current target, move target to last
                # qudit
                loc = list(op.location)
                loc = (
                    loc[0:op.gate.target_qubit]
                    + loc[(op.gate.target_qubit + 1):]
                    + [loc[op.gate.target_qubit]]
                )
                locations.append(CircuitLocation(loc))

        if len(ops) > 0:
            # Do a bulk QSDs -> circs
            if self.decompose_twice:
                circs = [
                    MGDPass.decompose_mpx_two_levels(
                        isinstance(op.gate, MPRYGate),
                        op.params,
                        op.num_qudits,
                    ) for op in ops
                ]
            else:
                circs = [
                    MGDPass.decompose_mpx_one_level(
                        isinstance(op.gate, MPRYGate),
                        op.params,
                        op.num_qudits,
                    ) for op in ops
                ]
            circ_gates = [CircuitGate(x) for x in circs]
            circ_ops = [
                Operation(x, locations[i], x._circuit.params)
                for i, x in enumerate(circ_gates)
            ]
            circuit.batch_replace(pts, circ_ops)
            circuit.unfold_all()

        circuit.unfold_all()


class QSDPass(BasePass):
    """
    A pass performing one round of decomposition from the QSD algorithm.

    This decomposition takes each unitary of size n and decomposes it
    into a circuit with 4 VariableUnitaryGates of size n - 1 and 3 multiplexed
    rotation gates.

    Important: This pass runs on VariableUnitaryGates only.

    References:
        https://arxiv.org/pdf/quant-ph/0406176
    """

    def __init__(
        self,
        min_qudit_size: int = 4,
    ) -> None:
        """
        Construct a single level of the QSDPass.

        Args:
            min_qudit_size (int): Performs a decomposition on all gates
                with width > min_qudit_size
        """
        self.min_qudit_size = min_qudit_size

    @staticmethod
    def shift_down_unitary(
        num_qudits: int,
        end_qubits: int,
    ) -> PermutationMatrix:
        """Return the Permutation Matrix that shifts the qubits down by 1
        qubit."""
        top_qubits = num_qudits - end_qubits
        now_bottom_qubits = list(reversed(range(top_qubits)))
        now_top_qubits = list(range(num_qudits - end_qubits, num_qudits))
        final_qudits = now_top_qubits + now_bottom_qubits
        return PermutationMatrix.from_qubit_location(num_qudits, final_qudits)

    @staticmethod
    def shift_up_unitary(
        num_qudits: int,
        end_qubits: int,
    ) -> PermutationMatrix:
        """Return the Permutation Matrix that shifts the qubits down by 1
        qubit."""
        bottom_qubits = list(range(end_qubits))
        top_qubits = list(reversed(range(end_qubits, num_qudits)))
        final_qudits = top_qubits + bottom_qubits
        return PermutationMatrix.from_qubit_location(num_qudits, final_qudits)

    @staticmethod
    def create_unitary_gate(u: UnitaryMatrix) -> tuple[
        VariableUnitaryGate,
        RealVector,
    ]:
        """Create a VariableUnitaryGate from a UnitaryMatrix."""
        gate = VariableUnitaryGate(u.num_qudits)
        params = np.concatenate((np.real(u).flatten(), np.imag(u).flatten()))
        return gate, params

    @staticmethod
    def create_multiplexed_circ(
        us: list[UnitaryMatrix],
        select_qubits: list[int],
    ) -> Circuit:
        """
        Takes a list of 2 unitaries of size n. Returns a circuit that decomposes
        the unitaries into a circuit with 2 unitaries of size n-1 and a
        multiplexed controlled gate.

        Args:
            us (list[UnitaryMatrix]): The unitaries to decompose
            select_qubits (list[int]): The qubits to use as select qubits
            controlled_qubit (int): The qubit to use as the controlled qubit

        Returns:
            Circuit: The circuit that decomposes the unitaries

        Using this paper: https://arxiv.org/pdf/quant-ph/0406176.pdf. Thm 12
        """
        u1 = us[0]
        u2 = us[1]
        assert (u1.num_qudits == u2.num_qudits)
        all_qubits = list(range(len(select_qubits) + 1))
        # Use Schur Decomposition to split Us into V, D, and W matrices
        D_2, V = schur(u1._utry @ u2.dagger._utry)
        # D^2 will be diagonal since u1u2h is unitary
        D = np.sqrt(np.diag(np.diag(D_2)))
        # Calculate W @ U1
        left_mat = D @ V.conj().T @ u2._utry
        left_gate, left_params = QSDPass.create_unitary_gate(
            UnitaryMatrix(left_mat),
        )

        # Create Multi Controlled Z Gate
        z_params: RealVector = np.array(-2 * np.angle(np.diag(D)).flatten())
        z_gate = MPRZGate(len(all_qubits), u1.num_qudits)

        # Create right gate
        right_gate, right_params = QSDPass.create_unitary_gate(UnitaryMatrix(V))

        circ = Circuit(u1.num_qudits + 1)
        circ.append_gate(left_gate, CircuitLocation(select_qubits), left_params)
        circ.append_gate(z_gate, CircuitLocation(all_qubits), z_params)
        circ.append_gate(
            right_gate, CircuitLocation(
                select_qubits,
            ), right_params,
        )
        return circ

    @staticmethod
    def mod_unitaries(u: UnitaryMatrix) -> UnitaryMatrix:
        """Apply a permutation transform to the unitaries to the rest of the
        circuit."""
        shift_up = QSDPass.shift_up_unitary(u.num_qudits, u.num_qudits - 1)
        shift_down = QSDPass.shift_down_unitary(u.num_qudits, u.num_qudits - 1)
        return shift_up @ u @ shift_down

    @staticmethod
    def qsd(orig_u: UnitaryMatrix) -> Circuit:
        '''
        Perform the Quantum Shannon Decomposition on a unitary matrix.
        Args:
            orig_u (UnitaryMatrix): The unitary matrix to decompose

        Returns:
            Circuit: The circuit that decomposes the unitary
        '''

        # Shift the unitary qubits down by one
        u = QSDPass.mod_unitaries(orig_u)

        # Perform CS Decomp to solve for multiplexed unitaries and theta_y
        (u1, u2), theta_y, (v1h, v2h) = cossin(
            u._utry, p=u.shape[0] / 2, q=u.shape[1] / 2, separate=True,
        )
        assert (len(theta_y) == u.shape[0] / 2)

        # Create the multiplexed circuit
        # This generates 2 circuits that multipex U,V with an MPRY gate
        controlled_qubit = u.num_qudits - 1
        select_qubits = list(range(0, u.num_qudits - 1))
        all_qubits = list(range(u.num_qudits))
        circ_1 = QSDPass.create_multiplexed_circ(
            [
                UnitaryMatrix(v1h), UnitaryMatrix(v2h),
            ],
            select_qubits,
        )
        circ_2 = QSDPass.create_multiplexed_circ(
            [
                UnitaryMatrix(u1), UnitaryMatrix(u2),
            ],
            select_qubits,
        )
        gate_2 = MPRYGate(u.num_qudits, controlled_qubit)

        circ_1.append_gate(gate_2, CircuitLocation(all_qubits), 2 * theta_y)
        circ_1.append_circuit(
            circ_2, CircuitLocation(list(range(u.num_qudits))),
        )
        return circ_1

    @staticmethod
    def get_variable_unitary_pts(
        circuit: Circuit,
        min_qudit_size: int,
    ) -> tuple[list[UnitaryMatrix], list[CircuitPoint], list[CircuitLocation]]:
        """Get all VariableUnitary Gates in the circuit wider than
        `min_qudit_size` and return their unitaries, points, and locations."""
        unitaries: list[UnitaryMatrix] = []
        pts: list[CircuitPoint] = []
        locations: list[CircuitLocation] = []
        num_ops = 0
        all_ops = list(circuit.operations_with_cycles(reverse=True))

        # Gather all of the VariableUnitary unitaries
        for cyc, op in all_ops:
            if (
                op.num_qudits > min_qudit_size
                and isinstance(op.gate, VariableUnitaryGate)
            ):
                num_ops += 1
                unitaries.append(op.get_unitary())
                pts.append(CircuitPoint((cyc, op.location[0])))
                locations.append(op.location)

        return unitaries, pts, locations

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform a single pass of Quantum Shannon Decomposition on the
        circuit."""
        unitaries, pts, locations = QSDPass.get_variable_unitary_pts(
            circuit, self.min_qudit_size,
        )

        if len(unitaries) > 0:
            circs = await get_runtime().map(QSDPass.qsd, unitaries)
            circ_gates = [CircuitGate(x) for x in circs]
            circ_ops = [
                Operation(x, locations[i], x._circuit.params)
                for i, x in enumerate(circ_gates)
            ]
            circuit.batch_replace(pts, circ_ops)
            circuit.unfold_all()

        circuit.unfold_all()



"""This module implements the VariableUnitaryGate."""
from __future__ import annotations

from typing import Sequence

import numpy as np

from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryLike
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_valid_radixes


class VariableUnitaryGate(GeneralGate):
    """A Variable n-qudit unitary operator."""

    def __init__(self, num_qudits: int, radixes: Sequence[int] = []) -> None:
        """
        Creates an VariableUnitaryGate, defaulting to a qubit gate.

        Args:
            num_qudits (int): The number of qudits this gate acts on.

            radixes (Sequence[int]): The number of orthogonal
                states for each qudit. Defaults to qubits.
        """
        if num_qudits <= 0:
            raise ValueError('Expected positive integer, got %d' % num_qudits)

        if len(radixes) == 0:
            radixes = [2] * num_qudits

        if not is_valid_radixes(radixes, num_qudits):
            raise TypeError('Invalid radixes.')

        self._num_qudits = int(num_qudits)
        self._radixes = tuple(radixes)
        self._dim = int(np.prod(self.radixes))
        self.shape = (self.dim, self.dim)
        self._num_params = 2 * self.dim**2
        self._name = 'VariableUnitaryGate(%d, %s)' % (
            self.num_qudits, str(self.radixes),
        )

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """
        Return the unitary for this gate, see :class:`Unitary` for more.

        Note:
            Ideally, params form a unitary matrix when reshaped,
            however, params are unconstrained so we return the closest
            UnitaryMatrix to the given matrix.
        """
        self.check_parameters(params)
        mid = len(params) // 2
        real = np.array(params[:mid], dtype=np.complex128)
        imag = 1j * np.array(params[mid:], dtype=np.complex128)
        x = real + imag
        return UnitaryMatrix.closest_to(np.reshape(x, self.shape), self.radixes)

    def calc_params(self, utry: UnitaryLike) -> list[float]:
        """Return the parameters for this gate to implement `utry`"""
        if 2 * len(utry) ** 2 != self.num_params:
            raise ValueError('Mismatch in unitary and gate dimension.')
        x = np.reshape(utry, (self.num_params // 2,))
        return list(np.real(x)) + list(np.imag(x))

    @staticmethod
    def get_params(utry: UnitaryLike) -> RealVector:
        """Return the params for this gate, given a unitary matrix."""
        num_elems = len(utry) ** 2
        real = np.reshape(np.real(utry), num_elems)
        imag = np.reshape(np.imag(utry), num_elems)
        return np.concatenate([real, imag])

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, VariableUnitaryGate)
            and self.num_qudits == other.num_qudits
            and self.radixes == other.radixes
        )

    def __hash__(self) -> int:
        return hash((self.num_qudits, self.radixes))



"""
This package contains several modules related to cost functions.

CostFunctions map circuit parameters, represented as a vector of real numbers,
to a real cost value. They are either generated by CostFunctionGenerators or
constructed normally. CostFunctions are minimized by minimizers, which change
the parameters such that the cost value is best minimized. Minimization can be
used on its own with `circuit.minimize` or as a circuit instantiation method
with `circuit.instantiate`.
"""
from __future__ import annotations

from bqskit.ir.opt.cost.function import CostFunction
from bqskit.ir.opt.cost.functions import HilbertSchmidtCost
from bqskit.ir.opt.cost.functions import HilbertSchmidtCostGenerator
from bqskit.ir.opt.cost.functions import HilbertSchmidtResiduals
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator

__all__ = [
    'CostFunction',
    'CostFunctionGenerator',
    'HilbertSchmidtCost',
    'HilbertSchmidtCostGenerator',
    'HilbertSchmidtResiduals',
    'HilbertSchmidtResidualsGenerator',
]



"""This package implements necessary algorithms and objects for circuit template
instantiation and cost minimization."""
from __future__ import annotations

from bqskit.ir.opt.cost.function import CostFunction
from bqskit.ir.opt.cost.functions import HilbertSchmidtCost
from bqskit.ir.opt.cost.functions import HilbertSchmidtCostGenerator
from bqskit.ir.opt.cost.functions import HilbertSchmidtResiduals
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.ir.opt.instantiater import Instantiater
from bqskit.ir.opt.instantiaters import Minimization
from bqskit.ir.opt.instantiaters import QFactor
from bqskit.ir.opt.minimizer import Minimizer
from bqskit.ir.opt.minimizers import CeresMinimizer
from bqskit.ir.opt.minimizers import LBFGSMinimizer
from bqskit.ir.opt.minimizers import ScipyMinimizer
from bqskit.ir.opt.multistartgen import MultiStartGenerator
from bqskit.ir.opt.multistartgens import DiagonalStartGenerator
from bqskit.ir.opt.multistartgens import RandomStartGenerator

__all__ = [
    'Instantiater',
    'QFactor',
    'Minimization',
    'Minimizer',
    'LBFGSMinimizer',
    'CostFunction',
    'CostFunctionGenerator',
    'HilbertSchmidtCost',
    'HilbertSchmidtCostGenerator',
    'HilbertSchmidtResiduals',
    'HilbertSchmidtResidualsGenerator',
    'MultiStartGenerator',
    'DiagonalStartGenerator',
    'RandomStartGenerator',
    'CeresMinimizer',
    'ScipyMinimizer',
]
